#!/bin/bash

#-----------------------------LICENSE NOTICE-----------------------------------
#     __    __  ______  __________
#    / /   / / / / __ \/  _/ ____/
#   / /   / / / / / / // // /
#  / /___/ /_/ / /_/ // // /___
# /_____/\____/_____/___/\____/
# (break the loop)
#
# This file is part of Amstrad CPC game "LUDIC break the loop".
#
#  "LUDIC break the loop" is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  "LUDIC break the loop" is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http:#www.gnu.org/licenses/>.
#
#  Code, Concept, Story Copyright (C) 2019 Naad (@teopl)
#  Art, Music, Story Copyright (C) 2019 Branho (@brando464)
#
#  Contact: teopl.the.game@gmail.com
#
# Developed for the CPCRetrodev contest 2019 edition.
#
#------------------------------------------------------------------------------

export PATH=${PATH}:${CPCT_PATH}/tools/sdcc-3.6.8-r9946/bin
export PATH=${PATH}:${CPCT_PATH}/tools/2cdt/bin
export PATH=${PATH}:${CPCT_PATH}/tools/cpc2cdt/bin
export PATH=${PATH}:${CPCT_PATH}/tools/dskgen/bin
export PATH=${PATH}:${CPCT_PATH}/tools/hex2bin-2.0/bin
export PATH=${PATH}:${CPCT_PATH}/tools/iDSK-0.13/bin
export PATH=${PATH}:${CPCT_PATH}/tools/img2cpc/bin
export PATH=${PATH}:${CPCT_PATH}/tools/rgas-1.2.2
export PATH=${PATH}:${CPCT_PATH}/tools/winape
export PATH=${PATH}:${CPCT_PATH}/tools/zx7b/bin

# include
source $(dirname $0)/scripts/portable_commands.sh
source $(dirname $0)/scripts/bash_library.sh

# ###

find . -iname '*.sh' -exec perl -i -p -e 's/\r//g' {} \;

mkdir -p src/assets/binary
mkdir -p src/assets/binary_cpctelera
mkdir -p src/assets/binary_wincpctelera
mkdir -p src/assets/binary_compressed

# REMARK: always clean
#make clean
#rm -rf ./obj

# FIXME:
# slice titlescreen (must be mode0 160x200)
#convert "assets/images/titlescreen.png" -crop 4x4-0-0@ +repage +adjoin assets/images/titlescreen_%d.png
#rm "assets/images/titlescreen.png"

# generate source assets
cd scripts >/dev/null 2>&1
./convert.sh
EXIT=$?
cd - >/dev/null 2>&1

# bail?
if [ $EXIT -ne 0 ]; then
    exit 1
fi

# update memory buffer size
SIZE_SONG=$(cat src/teopl_sound.h |grep TEOPL_SOUND_LARGEST_SONG |perl -p -e 's/[^\d]//g')

if [ -f src/assets/binary_compressed/ai_talk_00.h ]; then
    SIZE_TALK1=$(cat src/assets/binary_compressed/ai_talk_00.h |grep 'size ' |perl -p -e 's/.*size [^\d]*(\d+)/\1/g')
    SIZE_TALK2=$(cat src/assets/binary_compressed/ai_talk_01.h |grep 'size ' |perl -p -e 's/.*size [^\d]*(\d+)/\1/g')
    SIZE_TALK3=$(cat src/assets/binary_compressed/ludic_talk_00.h |grep 'size ' |perl -p -e 's/.*size [^\d]*(\d+)/\1/g')
    SIZE_TALK4=$(cat src/assets/binary_compressed/ludic_talk_01.h |grep 'size ' |perl -p -e 's/.*size [^\d]*(\d+)/\1/g')
    SIZE_TALK_AI=$((SIZE_TALK1+SIZE_TALK2))
    SIZE_TALK_LUDIC=$((SIZE_TALK3+SIZE_TALK4))

    if [ $SIZE_TALK_AI -gt $SIZE_TALK_LUDIC ]; then
        SIZE_TALK=${SIZE_TALK_AI}
    else
        SIZE_TALK=${SIZE_TALK_LUDIC}
    fi
else    
    SIZE_TALK=504
    echo "${COLOR_YELLOW}*** MAKE: Check buffer: Make sure ${SIZE_TALK} is exact space for ai or ludic talk animations (which is larger) ${COLOR_NORMAL}"
fi

# calculate space for all assets
SIZE_SONG_AND_TALK=$((SIZE_SONG+SIZE_TALK))

SIZE_LEVEL_TEXTS_ENCODED_LINE=$(cat src/teopl_level.h |grep '#define TEOPL_LEVEL_TEXTS_ENCODED_LINE' |perl -p -e 's/[^\d]//g')
SIZE_LEVEL_LEVEL_MAX_NUM_TEXTS=$(cat src/teopl_level.h |grep '#define TEOPL_LEVEL_MAX_NUM_TEXTS' |perl -p -e 's/[^\d]//g')
SIZE_LEVEL_AND_TEXT=$(bc <<< "325 + ${SIZE_LEVEL_LEVEL_MAX_NUM_TEXTS} * (${SIZE_LEVEL_TEXTS_ENCODED_LINE} + 3)")

SIZE_ALL_ASSETS=$((SIZE_SONG_AND_TALK+SIZE_LEVEL_AND_TEXT))

if [ -f src/assets/binary_compressed/texts_levels.h ]; then
    SIZE_TEXT=$(cat src/assets/binary_compressed/texts_levels.h |grep 'size ' |perl -p -e 's/.*size [^\d]*(\d+)/\1/g')    
else
    SIZE_TEXT=1642
    echo "${COLOR_RED}*** MAKE: Make sure that text size (${SIZE_TEXT}) matches the on in texts_levels.h! ${COLOR_NORMAL}"
fi
SIZE_TEXT=$((SIZE_TEXT+1))

# buffer is used like this:
# [SONG] [LEVEL] [TEXT] [FREE]? [TALK]
# or
# [ T      E       X       T       S ]
# ###################################

echo "${COLOR_YELLOW}*** MAKE: Check buffer: SIZE_TEXT(plus one)=${SIZE_TEXT} SIZE_SONG_AND_TALK=${SIZE_SONG_AND_TALK} SIZE_LEVEL_AND_TEXT=${SIZE_LEVEL_AND_TEXT} ${COLOR_NORMAL}"
echo "${COLOR_YELLOW}*** MAKE: Check buffer: SIZE_ALL_ASSETS=${SIZE_ALL_ASSETS} ${COLOR_NORMAL}"

SIZE_BUFFER_FINAL=0
SIZE_REMAINING=0
if [ $SIZE_TEXT -gt $SIZE_ALL_ASSETS ]; then
    SIZE_REMAINING=$((SIZE_TEXT-SIZE_ALL_ASSETS))
    SIZE_BUFFER_FINAL=${SIZE_TEXT}
else
    SIZE_BUFFER_FINAL=${SIZE_ALL_ASSETS}
fi
echo "${COLOR_YELLOW}*** MAKE: Check buffer: SIZE_BUFFER_FINAL=${SIZE_BUFFER_FINAL} ${COLOR_NORMAL}"

if [ $SIZE_REMAINING -eq 0 ]; then
    echo "${COLOR_YELLOW}*** MAKE: Check buffer: After the song (after ${SIZE_SONG} bytes) there is no space before 'talk' animations on the end!${COLOR_NORMAL}"
else
    echo "${COLOR_RED}*** MAKE: Check buffer: After the song (after ${SIZE_SONG} bytes) there is free ${SIZE_REMAINING} bytes before 'talk' animations on the end!${COLOR_NORMAL}"
fi

# [64, 65]  or [0x40, 0x41] are for u16 value for firmware address returned by cpct_disableFirmware()
# [66, 255] or [0x42, 0xFF] is for text constants like "START", "PASSWORD", ...
# [256, 511] or [0x100, 0x1FF] is where the aligned masktable must be (check g_teopl_video_masktable)
CODELOC=$((512+SIZE_BUFFER_FINAL+1))
CODELOC_HEX=$(echo "obase=16;ibase=10;${CODELOC}" |bc)

# set CODELOC and BUFFER SIZE
perl -i -p -e "s/Z80CODELOC := .*/Z80CODELOC := 0x${CODELOC_HEX}/g" cfg/build_config.mk
perl -i -p -e "s/  TEOPL_POOL_MAX_ANYTHING.*/  TEOPL_POOL_MAX_ANYTHING ${SIZE_BUFFER_FINAL}/g" src/teopl_pool.h

# ---

# convert to unix new lines (avoid false modified git status)
# FIXME: do this better - we don't have flip on mac
#cd .. >/dev/null 2>&1
#for f in $(git ls-files -m); do flip -u "${f}" >/dev/null 2>&1; done
#cd - >/dev/null 2>&1

# remove backup files
find . -type f -iname '*.bak' -exec rm {} +

# code build
esc=$(printf '\033')
rm build.log teopl.dsk >/dev/null 2>&1
make all 2>&1 |${CMD_SED} "s,error,${esc}[31m&${esc}[0m,gi" |${CMD_SED} "s,warning,${esc}[35m&${esc}[0m,gi" |tee build.log

# replace memory address for tcpc
#NEW_HIGHEST_ADDRESS_HEX=$(cat build.log |grep 'Highest address' |perl -p -e 's/^.*=[ 0]+//g')
#NEW_HIGHEST_ADDRESS_DEC=$(echo "obase=10;ibase=16;${NEW_HIGHEST_ADDRESS_HEX}" |bc)
#OLD_HIGHEST_ADDRESS_DEC="$(cat assets/LOADER.BAS |grep MEMORY |perl -p -e 's/.*MEMORY\s(\d+).*/\1/g')"

#HIGHEST_ADDRESS_DEC=$(echo "obase=10;ibase=10;${OLD_HIGHEST_ADDRESS_DEC}" |bc)
#HIGHEST_ADDRESS_PLUS_ONE_HEX=$(echo "obase=16;ibase=10;${OLD_HIGHEST_ADDRESS_DEC}+1" |bc)

#if [ "$NEW_HIGHEST_ADDRESS_HEX" != "" ] && [ "$NEW_HIGHEST_ADDRESS_DEC" != "$OLD_HIGHEST_ADDRESS_DEC" ]; then

# FIXME: remove
#if false; then
#
#    HIGHEST_ADDRESS_DEC=$(echo "obase=10;ibase=16;${NEW_HIGHEST_ADDRESS_HEX}" |bc)
#    HIGHEST_ADDRESS_PLUS_ONE_HEX=$(echo "obase=16;ibase=16;${NEW_HIGHEST_ADDRESS_HEX}+1" |bc)
#
#    # align
#    #HIGHEST_ADDRESS_PLUS_ONE_HEX=$(echo "obase=16;ibase=16;${HIGHEST_ADDRESS_PLUS_ONE_HEX}+${HIGHEST_ADDRESS_PLUS_ONE_HEX}%2" |bc)
#
#    perl -i -p -e "s/MEMORY.*/MEMORY ${HIGHEST_ADDRESS_DEC}/g" ./assets/LOADER.BAS
#    perl -i -p -e "s/^org\s.*&.*/org &${HIGHEST_ADDRESS_PLUS_ONE_HEX}/g" ./src_extern/tcpc_entrypoint.asm
#    perl -i -p -e "s/#define\s+TEOPL_TCPC_ADDRESS\s+.*/#define TEOPL_TCPC_ADDRESS #0x${HIGHEST_ADDRESS_PLUS_ONE_HEX}/g" ./src/teopl_tcpc.h
#
#    #ADDRESS_SOUND_BUF_HEX=$(cat ./obj/teopl_sound.rst |grep g_song_data:: |perl -p -e 's/\s+([A-F0-9]+).*/\1/g') >/dev/null 2>&1
#    #echo "${COLOR_RED}*** MAKE: Highest address updated, need to rebuild... $OLD_HIGHEST_ADDRESS_DEC => $NEW_HIGHEST_ADDRESS_DEC (sound=${ADDRESS_SOUND_BUF_HEX}) ${COLOR_NORMAL}"
#    echo "${COLOR_RED}*** MAKE: Highest address updated, need to rebuild... $OLD_HIGHEST_ADDRESS_DEC => $NEW_HIGHEST_ADDRESS_DEC ${COLOR_NORMAL}"
#    #rm -rf ./src/assets/sound
#
#    exit 1
#fi

# depending on where larges song ends, prepare tcpc asm code
# TCPC_LOC_DEC=$(cat src/teopl_sound.h |grep TEOPL_SOUND_LARGEST_SONG |perl -p -e "s/[^\d]*(\d+)/\1/g")
# TCPC_LOC_HEX=$(echo "obase=16;ibase=10;64+2+${TCPC_LOC_DEC}+1" |bc) # 64+2 bytes is where buffer starts
#perl -i -p -e "s/org &.*/org &${TCPC_LOC_HEX}/g" ./src_extern/tcpc_entrypoint.asm
#perl -i -p -e "s/#define\s+TEOPL_TCPC_ADDRESS\s+.*/#define TEOPL_TCPC_ADDRESS #0x${TCPC_LOC_HEX}/g" ./src/teopl_tcpc.h

# prepare tcpc asm code to be placed at the buffer start (over the song!)
TCPC_LOC_HEX="200"
perl -i -p -e "s/org &.*/org &${TCPC_LOC_HEX}/g" ./src_extern/tcpc_entrypoint.asm
perl -i -p -e "s/#define\s+TEOPL_TCPC_ADDRESS\s+.*/#define TEOPL_TCPC_ADDRESS #0x${TCPC_LOC_HEX}/g" ./src/teopl_tcpc.h

# compile external sources
# FIXME: pasmo doesn't work with wine on macos
if [[ ! "$OSTYPE" == "darwin"* ]]; then

	cd src_extern
	rm tcpc.bin tcpc.c tcpc.h >/dev/null 2>&1
	${CMD_WINE} ./pasmo.exe tcpc_entrypoint.asm tcpc.bin

    # compress tcpc code to be extracted to TCPC_LOC_HEX when needed
    ../scripts/compress_binary.sh EXO2 CPC tcpc.bin tcpc.c
    mv tcpc.c ../src
    mv tcpc.h ../src

	if [ $? -ne 0 ]; then
	    echo "${COLOR_RED}*** MAKE: extern source build failed!${COLOR_NORMAL}"
	    exit 1
	fi

	# FIXME: build server on all platforms? currently links wsock32 in Makefile
	#make clean >/dev/null 2>&1
	#make >/dev/null 2>&1

	#if [ $? -ne 0 ]; then
	#    echo "${COLOR_RED}*** MAKE: extern source build failed!${COLOR_NORMAL}"
	#    exit 1
	#fi
	cd -
fi

# insert additional files in dsk
# FIXME: do this better, we don't have flip on mac. This is important, do not skip!
flip -m ./assets/LOADER.BAS
${CMD_WINE} ./scripts/iDSK.exe teopl.dsk -i ./src/assets/binary_cpctelera/titlescreenfull_cpctelera.bin -t 1 -c C000 -f >/dev/null 2>&1
#${CMD_WINE} ./scripts/iDSK.exe teopl.dsk -i ./src_extern/tcpc.bin -t 1 -c ${HIGHEST_ADDRESS_PLUS_ONE_HEX} -f >/dev/null 2>&1
#${CMD_WINE} ./scripts/iDSK.exe teopl.dsk -i ./src/assets/sound/arkos2/sound_all.bin -t 1 -c 40 -f >/dev/null 2>&1
${CMD_WINE} ./scripts/iDSK.exe teopl.dsk -i ./assets/LOADER.BAS -t 0 -f >/dev/null 2>&1
${CMD_WINE} ./scripts/iDSK.exe teopl.dsk -i ./assets/TEOPL.BAS -t 1 -c 170 -f >/dev/null 2>&1
#${CMD_WINE} ./scripts/managedsk/ManageDsk -Lteopl.dsk -ALOADER.BAS -Steopl.dsk >/dev/null 2>&1
#${CMD_WINE} ./scripts/CPCDiskXP.exe -File LOADER.BAS -AddAmsdosHeader 170 -AddToExistingDsk teopl.dsk

# after make, prepare compressed binary for wincpctelera build
cd scripts >/dev/null 2>&1
./convert_binary.sh "../src/assets/binary" "../src/assets/binary_compressed_wincpctelera/common"
./convert_binary.sh "../src/assets/binary_wincpctelera" "../src/assets/binary_compressed_wincpctelera/only"

if [ $? -ne 0 ]; then
    exit 1
fi

EXIT=$?

cd - >/dev/null 2>&1
