#!/bin/bash

#-----------------------------LICENSE NOTICE-----------------------------------
#     __    __  ______  __________
#    / /   / / / / __ \/  _/ ____/
#   / /   / / / / / / // // /
#  / /___/ /_/ / /_/ // // /___
# /_____/\____/_____/___/\____/
# (break the loop)
#
# This file is part of Amstrad CPC game "LUDIC break the loop".
#
#  "LUDIC break the loop" is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  "LUDIC break the loop" is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http:#www.gnu.org/licenses/>.
#
#  Code, Concept, Story Copyright (C) 2019 Naad (@teopl)
#  Art, Music, Story Copyright (C) 2019 Branho (@brando464)
#
#  Contact: teopl.the.game@gmail.com
#
# Developed for the CPCRetrodev contest 2019 edition.
#
#------------------------------------------------------------------------------

# include
source $(dirname $0)/bash_library.sh

if (( $# != 3 )); then
    echo "${COLOR_RED}*** SOUND: Illegal number of parameters!${COLOR_NORMAL}"
    exit 1
fi

# ###

FOLDER_ASSETS_INPUT=$1
FOLDER_ASSETS_OUTPUT1=$2
FOLDER_ASSETS_OUTPUT2=$3

 # bail?
 if [ -d $FOLDER_ASSETS_OUTPUT1 ]; then
    printf "*** SND: No need to regenerate...\n\n"
    exit 0
fi

mkdir -p ${FOLDER_ASSETS_OUTPUT1}

rm -rf ${FOLDER_ASSETS_OUTPUT2}
mkdir -p ${FOLDER_ASSETS_OUTPUT2}

# contain player and effects (should be same for all songs)
FILE_COMMON="${FOLDER_ASSETS_OUTPUT2}/common.s"

ERROR="no"

# ###

# determine address for sound buffer (or set mock one)
#ADDRESS_SOUND_BUF_HEX_MOCK="4000"
#if [[ -f ../obj/teopl_sound.rst ]]; then
    
    #if [[ -f ../src/ADDRESS_SOUND_BUF_HEX ]]; then 
    #    ADDRESS_SOUND_BUF_HEX=$(cat ../src/ADDRESS_SOUND_BUF_HEX)
    #else
        #ADDRESS_SOUND_BUF_HEX=$(cat ../obj/teopl_sound.rst |grep g_song_data:: |perl -p -e 's/\s+([A-F0-9]+).*/\1/g') >/dev/null 2>&1
    #fi

    #echo "${COLOR_GREEN}*** SND: sound buffer found at: ${ADDRESS_SOUND_BUF_HEX} ${COLOR_NORMAL}"
#else        
    #ADDRESS_SOUND_BUF_HEX="${ADDRESS_SOUND_BUF_HEX_MOCK}" >/dev/null 2>&1
#fi

#ADDRESS_SOUND_BUF_HEX=$(cat ../src/teopl_tcpc.h |grep TEOPL_TCPC_ADDRESS |perl -p -e 's/.*0x([0-9a-fA-F]+).*/\1/g')
ADDRESS_SOUND_BUF_HEX="200"
#echo "${COLOR_GREEN}*** SND: sound buffer found at: ${ADDRESS_SOUND_BUF_HEX} ${COLOR_NORMAL}"

# 1. convert sfxs
# ###

# REMARK: support only one sfx file (hard codded name)

SFX_INPUT="${FOLDER_ASSETS_INPUT}/sfx_main.aks"
SFX_NAME=$(basename $SFX_INPUT .aks)

echo "*** SND: check '${SFX_INPUT}'..."

SFX_OUTPUT="${FOLDER_ASSETS_OUTPUT1}/${SFX_NAME}.asm"
SFX_OUTPUT_CONFIG="${FOLDER_ASSETS_OUTPUT1}/${SFX_NAME}_playerconfig.asm"
${CMD_WINE} ./arkos2/tools/SongToSoundEffects.exe --exportPlayerConfig --labelPrefix "${SFX_NAME}_" ${SFX_INPUT} ${SFX_OUTPUT} >/dev/null 2>&1

echo "*** SND: generated '${SFX_OUTPUT}'"
echo "*** SND: generated '${SFX_OUTPUT_CONFIG}'"

echo

# 2. convert songs
# ###

PLAYER_ID="LW"

if [ "$PLAYER_ID" == "LW" ]; then
    SONG_CONVERTER="SongToLightweight.exe"
    PLAYER="PlayerLightweight"
#elif [ "$PLAYER_ID" == "AKG" ]; then
#    SONG_CONVERTER="SongToAkg.exe"
#    PLAYER="PlayerAkg"
else
    ERROR="*** SND: ERROR: Invalid player id ***"
    break
fi

# ###

for f in $(find ${FOLDER_ASSETS_INPUT} -type f -iname 'song*.aks'); do

    SONG_NAME=$(basename $f .aks)
    SONG_INPUT=${f}
    
    echo "*** SND: check '${SONG_INPUT}'..."

    # REMARK: all songs, sfx and the player
    ARKOS2_FILE_SOUND_ALL="${FOLDER_ASSETS_OUTPUT1}/sound_all_${SONG_NAME}.asm"
    ARKOS2_FILE_SOUND_ALL_BIN="${FOLDER_ASSETS_OUTPUT1}/sound_all_${SONG_NAME}.bin"

    rm -rf ${ARKOS2_FILE_SOUND_ALL} >/dev/null 2>&1
    touch ${ARKOS2_FILE_SOUND_ALL}

    printf "; Generated file: all of music, sfx and the player\n\n" >> ${ARKOS2_FILE_SOUND_ALL}

    printf "; This is the music, and its config file.\n" >> ${ARKOS2_FILE_SOUND_ALL}

    # convert song
    SONG_OUTPUT="${FOLDER_ASSETS_OUTPUT1}/${SONG_NAME}.asm"
    ${CMD_WINE} ./arkos2/tools/${SONG_CONVERTER} --exportPlayerConfig --labelPrefix "${SONG_NAME}_" ${SONG_INPUT} ${SONG_OUTPUT} >/dev/null 2>&1

    # include song
    # ###

    printf "include \"$(basename ${SONG_OUTPUT})\"\n" >> ${ARKOS2_FILE_SOUND_ALL}
    printf "include \"$(basename ${SONG_NAME}_playerconfig.asm)\" ;Optional.\n" >> ${ARKOS2_FILE_SOUND_ALL}

    # include sfx
    # ###

    printf "\n; This is the sfxs, and its config file.\n" >> ${ARKOS2_FILE_SOUND_ALL}

    printf "include \"$(basename ${SFX_OUTPUT})\"\n" >> ${ARKOS2_FILE_SOUND_ALL}
    printf "include \"$(basename ${SFX_NAME}_playerconfig.asm)\" ;Optional.\n" >> ${ARKOS2_FILE_SOUND_ALL}

    # include setup and PLAYER
    # ###

    printf "\n; What hardware? Uncomment the right one.\n" >> ${ARKOS2_FILE_SOUND_ALL}
    printf "PLY_${PLAYER_ID}_HARDWARE_CPC = 1\n" >> ${ARKOS2_FILE_SOUND_ALL}
    printf "; PLY_${PLAYER_ID}_HARDWARE_MSX = 1\n" >> ${ARKOS2_FILE_SOUND_ALL}
    printf ";PLY_${PLAYER_ID}_HARDWARE_SPECTRUM = 1\n" >> ${ARKOS2_FILE_SOUND_ALL}
    printf ";PLY_${PLAYER_ID}_HARDWARE_PENTAGON = 1\n" >> ${ARKOS2_FILE_SOUND_ALL}

    printf "\n; Comment/delete this line if not using sound effects.\n" >> ${ARKOS2_FILE_SOUND_ALL}

    # include sfx
    printf "PLY_${PLAYER_ID}_MANAGE_SOUND_EFFECTS = 1\n\n" >> ${ARKOS2_FILE_SOUND_ALL}

    cp "${FOLDER_ASSETS_INPUT}/${PLAYER}.asm" ${FOLDER_ASSETS_OUTPUT1}
    cp "${FOLDER_ASSETS_INPUT}/${PLAYER}_SoundEffects.asm" ${FOLDER_ASSETS_OUTPUT1}

    printf "; This is the player.\n" >> ${ARKOS2_FILE_SOUND_ALL}
    printf "include \"${PLAYER}.asm\"\n" >> ${ARKOS2_FILE_SOUND_ALL}

    # generate SDCC compatible player and songs. Then split them extracting songs and compress songs.
    # ###

    # build binary
    SOUND_ALL_OUT="$(dirname ${ARKOS2_FILE_SOUND_ALL_BIN})/$(basename ${ARKOS2_FILE_SOUND_ALL_BIN} .bin)"
    SOUND_ALL_SDCC_ASM="$(dirname ${ARKOS2_FILE_SOUND_ALL})/$(basename ${ARKOS2_FILE_SOUND_ALL} .asm).s"
    SOUND_ALL_SDCC_SYM="$(dirname ${ARKOS2_FILE_SOUND_ALL})/$(basename ${ARKOS2_FILE_SOUND_ALL} .asm).sym"
    ${CMD_WINE} ./rasm/rasm.exe ${ARKOS2_FILE_SOUND_ALL} -o ${SOUND_ALL_OUT} -s -sl >/dev/null 2>&1

    # convert to sdcc asm (.s)
    ${CMD_WINE} ./arkos2/tools/Disark.exe ${ARKOS2_FILE_SOUND_ALL_BIN} ${SOUND_ALL_SDCC_ASM} --symbolFile ${SOUND_ALL_SDCC_SYM} --sourceProfile sdcc >/dev/null 2>&1
    rm ${ARKOS2_FILE_SOUND_ALL_BIN} >/dev/null 2>&1
    
    # extract SONG and COMMON data
    SONG_SOURCE=${FOLDER_ASSETS_OUTPUT1}/${SONG_NAME}.s
    FILE_COMMON_NEW=${FILE_COMMON}.NEW
    perl -ne 'if(/SFX_MAIN_SOUNDEFFECTS:/){$a=1} ; $a==1 ? print STDERR : print STDOUT;' ${SOUND_ALL_SDCC_ASM} >${SONG_SOURCE} 2>${FILE_COMMON_NEW}

    if [[ ! -e "${FILE_COMMON}" ]]; then
        #diff ${FILE_COMMON_NEW} ${FILE_COMMON} >/dev/null || ERROR="*** SND: ERROR: Songs don't have same common code. Check if they are really extracted from one .aks with subsongs. ***"
        mv ${FILE_COMMON_NEW} ${FILE_COMMON}
        echo "*** SND: extracted '${FILE_COMMON}'"
    else
        SIZE_FILE_COMMON=$(wc -c ${FILE_COMMON} |cut -d' ' -f1)
        SIZE_FILE_COMMON_NEW=$(wc -c ${FILE_COMMON_NEW} |cut -d' ' -f1)

        if [[ ${SIZE_FILE_COMMON_NEW} -gt ${SIZE_FILE_COMMON} ]]; then
            mv ${FILE_COMMON_NEW} ${FILE_COMMON}
            echo "${COLOR_YELLOW}*** SND: extracted AGAIN '${FILE_COMMON}' - found larger common data - problem in .aks files?${COLOR_NORMAL}"
        fi
    fi

    rm -rf ${FILE_COMMON_NEW} >/dev/null 2>&1

    echo "*** SND: extracted '${SONG_SOURCE}'"

    # compress songs
    ${CMD_WINE} sdasz80.exe -l -o -s ${SONG_NAME}.rel ${SONG_SOURCE} >/dev/null 2>&1
    ${CMD_WINE} sdcc.exe -mz80 --no-std-crt0 -Wl-u --code-loc "0x${ADDRESS_SOUND_BUF_HEX}" --data-loc 0 ${SONG_NAME}.rel -o "${SONG_NAME}.ihx" >/dev/null 2>&1
    ${CMD_WINE} hex2bin.exe -p 00 "${SONG_NAME}.ihx" >/dev/null 2>&1
    ./compress_binary.sh EXO2 CPC ${SONG_NAME}.bin ${SONG_NAME}.c
    mv ${SONG_NAME}.c ${FOLDER_ASSETS_OUTPUT2}
    mv ${SONG_NAME}.h ${FOLDER_ASSETS_OUTPUT2}

    # remove files
    rm -rf ${SONG_NAME}.* >/dev/null 2>&1
    rm -rf ${SONG_SOURCE} >/dev/null 2>&1
    rm -rf ${FOLDER_ASSETS_OUTPUT1}/*.sym >/dev/null 2>&1
    rm -rf ${FOLDER_ASSETS_OUTPUT1}/sound_*.s >/dev/null 2>&1

    echo "*** SND: compressed to song '${SONG_NAME}.c'"
    echo
done

rm -rf ${FOLDER_ASSETS_OUTPUT1}/*.asm >/dev/null 2>&1

#if [ "${ADDRESS_SOUND_BUF_HEX}" == "${ADDRESS_SOUND_BUF_HEX_MOCK}" ]; then
#    echo "${COLOR_RED}*** SND: Must rebuild to include correct address for sound buffer!${COLOR_NORMAL}"
#    echo
#    rm -rf ${FOLDER_ASSETS_OUTPUT1}
#fi

# ###

# error?
if [[ "$ERROR" != 'no' ]]; then
    echo "${COLOR_RED}${ERROR}${COLOR_NORMAL}"
    rm -rf ${SONG_OUTPUT} ${SFX_OUTPUT} >/dev/null 2>&1
    exit 1
fi

# set min buffer size to fit all songs
LARGEST_SONG_SIZE=$(cat ${FOLDER_ASSETS_OUTPUT2}/*.h |grep '_size ' |cut -d' ' -f3 |sort -rn |head -n1)

# REMARK hack due to memory overlapping (720 byes is for menu_text_size)
if [ $LARGEST_SONG_SIZE -lt 720 ]; then
    LARGEST_SONG_SIZE=720
    echo "${COLOR_RED}*** SND: largest song buffer EXTENDED due to hotfix: ${LARGEST_SONG_SIZE} ${COLOR_NORMAL}"
else
    echo "${COLOR_GREEN}*** SND: largest song buffer: ${LARGEST_SONG_SIZE} ${COLOR_NORMAL}"
fi

perl -i -p -e "s/TEOPL_SOUND_LARGEST_SONG.*/TEOPL_SOUND_LARGEST_SONG ${LARGEST_SONG_SIZE}/g" ../src/teopl_sound.h

# output file
echo "${COLOR_LIGHT_GREEN}*** SND: REGENERATED!${COLOR_NORMAL}"

echo
