#!/bin/bash

#-----------------------------LICENSE NOTICE-----------------------------------
#     __    __  ______  __________
#    / /   / / / / __ \/  _/ ____/
#   / /   / / / / / / // // /
#  / /___/ /_/ / /_/ // // /___
# /_____/\____/_____/___/\____/
# (break the loop)
#
# This file is part of Amstrad CPC game "LUDIC break the loop".
#
#  "LUDIC break the loop" is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  "LUDIC break the loop" is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http:#www.gnu.org/licenses/>.
#
#  Code, Concept, Story Copyright (C) 2019 Naad (@teopl)
#  Art, Music, Story Copyright (C) 2019 Branho (@brando464)
#
#  Contact: teopl.the.game@gmail.com
#
# Developed for the CPCRetrodev contest 2019 edition.
#
#------------------------------------------------------------------------------

# include
source $(dirname $0)/bash_library.sh

if (( $# != 2 )); then
    echo "${COLOR_RED}*** BIN: Illegal number of parameters!${COLOR_NORMAL}"
    exit 1
fi

max()
{
    local m="$1"
    for n in "$@"
    do
        [ "$n" -gt "$m" ] && m="$n"
    done
    echo "$m"
}

# ---

FOLDER_ASSETS_IN=$1
FOLDER_ASSETS_OUT=$2

 # bail?
 if [ -d "$FOLDER_ASSETS_OUT" ]; then
    printf "*** BIN: No need to regenerate...\n\n"
    exit 0
fi

mkdir -p $FOLDER_ASSETS_OUT

ERROR="no"

for f in $(find ${FOLDER_ASSETS_IN} -type f -iname '*.bin'); do

    FILENAME=$(basename -- "$f")
    #EXTENSION="${filename##*.}"
    BIN_NAME="${FILENAME%.*}"

    echo "*** BIN: check '${FILENAME}'..."

    # ---

    BIN_OUTPUT_BASENAME="${BIN_NAME}"
    BIN_OUTPUT_COMPRESSED="${FOLDER_ASSETS_IN}/${BIN_OUTPUT_BASENAME}.compressed"
    BIN_OUTPUT_C="${FOLDER_ASSETS_OUT}/${BIN_OUTPUT_BASENAME}.c"
    BIN_OUTPUT_H="${FOLDER_ASSETS_OUT}/${BIN_OUTPUT_BASENAME}.h"

    # compress binary file
    ${CMD_WINE} ./zx7/zx7.exe -f ${f} ${BIN_OUTPUT_COMPRESSED} >/dev/null 2>&1

    # convert compressed binary to C array
    cpct_bin2c ${BIN_OUTPUT_COMPRESSED} -an ${BIN_OUTPUT_BASENAME} >${BIN_OUTPUT_C}
    rm -rf ${BIN_OUTPUT_COMPRESSED} >/dev/null 2>&1

    # create .c file
    # ---

    # use u8 type
    #perl -i -p -e 's/const unsigned char/const u8/g' ${BIN_OUTPUT_C}

    # add to beggining of file
    ${CMD_SED} -i "1s/^/#ifdef WINCPCTELERA\n/" ${BIN_OUTPUT_C}
    ${CMD_SED} -i "1s/^/#include <cpctelera.h>\n\n/" ${BIN_OUTPUT_C}
    ${CMD_SED} -i "1s/^/\/\/ REMARK: !!! DO NOT EDIT !!! This file is auto-generated.\n\n/" ${BIN_OUTPUT_C}

    # add to end of file
    printf "\n#endif\n" >>${BIN_OUTPUT_C}

    echo "*** BIN: generated '${BIN_OUTPUT_C}'"

    # create .h file
    # ---

    cat ${BIN_OUTPUT_C} |grep const| perl -p -e 's/const\s+unsigned\s+char\s+(\w+)\s*\[(\d+)\].*/#define \1_size_z \2 \nextern const unsigned char \1\[\2];\n#define \1_end (\1 \+ \2 - 1)\n/g' >${BIN_OUTPUT_H}

    SIZE_UNZIP=$(wc -c ${f} |cut -d' ' -f1)
    printf "#define ${BIN_OUTPUT_BASENAME}_size ${SIZE_UNZIP}\n" >>${BIN_OUTPUT_H}

    ${CMD_SED} -i "1s/^/\/\/ REMARK: !!! DO NOT EDIT !!! This file is auto-generated.\n/" ${BIN_OUTPUT_H}

    echo "*** BIN: generated '${BIN_OUTPUT_H}'"

done

# error?
if [[ "$ERROR" != 'no' ]]; then
    echo "${COLOR_RED}${ERROR}${COLOR_NORMAL}"
    rm -rf ${BIN_OUTPUT_C} >/dev/null 2>&1
    exit 1
fi

# output file
echo "${COLOR_LIGHT_GREEN}*** BIN: REGENERATED!${COLOR_NORMAL}"

echo
