#!/bin/bash

#-----------------------------LICENSE NOTICE-----------------------------------
#     __    __  ______  __________
#    / /   / / / / __ \/  _/ ____/
#   / /   / / / / / / // // /
#  / /___/ /_/ / /_/ // // /___
# /_____/\____/_____/___/\____/
# (break the loop)
#
# This file is part of Amstrad CPC game "LUDIC break the loop".
#
#  "LUDIC break the loop" is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  "LUDIC break the loop" is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http:#www.gnu.org/licenses/>.
#
#  Code, Concept, Story Copyright (C) 2019 Naad (@teopl)
#  Art, Music, Story Copyright (C) 2019 Branho (@brando464)
#
#  Contact: teopl.the.game@gmail.com
#
# Developed for the CPCRetrodev contest 2019 edition.
#
#------------------------------------------------------------------------------

# include
source $(dirname $0)/bash_library.sh

if (( $# != 2 )); then
    echo "${COLOR_RED}*** LVL: Illegal number of parameters!${COLOR_NORMAL}"
    exit 1
fi

max()
{
    local m="$1"
    for n in "$@"
    do
        [ "$n" -gt "$m" ] && m="$n"
    done
    echo "$m"
}

# #######################################################################################################################################

# ---

FOLDER_ASSETS_INPUT=$1
FOLDER_ASSETS_OUTPUT=$2

 # bail?
 if [ -d "$FOLDER_ASSETS_OUTPUT" ]; then
    printf "*** LVL: No need to regenerate...\n\n"
    exit 0
fi

rm -rf ${FILE_INTERNET_LEVELS_C} >/dev/null 2>&1

mkdir -p $FOLDER_ASSETS_OUTPUT

ERROR="no"

for f in $(find ${FOLDER_ASSETS_INPUT} -type f -iname '*.rgas'); do

    FILENAME=$(basename -- "$f")
    #EXTENSION="${filename##*.}"
    LEVELSET_NAME="${FILENAME%.*}"

    LEVELSET_INPUT="${FOLDER_ASSETS_INPUT}/${LEVELSET_NAME}.rgas"

    echo "*** LVL: check '${LEVELSET_INPUT}'..."

    HAS_LEVELS="$(grep '_aspectRatio' -i ${LEVELSET_INPUT} |tr -d '\0')"

    if [ "$HAS_LEVELS" == "" ]; then
        ERROR="*** GFX: ERROR: There are no levels in RGAS file ***"
        printf "'%s'\n" $HAS_LEVELS
        break
    fi

    # ---

    # check input file format
    INPUT_FORMAT1_OK="$(cat ${LEVELSET_INPUT} |grep '"RGAS.PlatformProviders.AmstradSpriteSet+AmstradCPCRSLibCArrayFormatter,')"
    INPUT_FORMAT2_OK="$(cat ${LEVELSET_INPUT} |grep '"OutputSpriteDimensions": true')"
    INPUT_FORMAT3_OK="$(cat ${LEVELSET_INPUT} |grep  \"Width\": |uniq  |${CMD_WC} -l)$(cat ${LEVELSET_INPUT} |grep \"Height\": |uniq  |${CMD_WC} -l)"
    INPUT_FORMAT3_OK=$(echo $INPUT_FORMAT3_OK | tr -d ' ')

    if [ "$INPUT_FORMAT1_OK" == "" ] || [ "$INPUT_FORMAT2_OK" == "" ]; then
        ERROR="*** LVL: ERROR: Make sure to select 'CPCTelera/SDCC C array' and 'Output dimensions for sprite' in RGAS ***"
        break
    fi
    if [ "$INPUT_FORMAT3_OK" != "11" ]; then
        ERROR="*** LVL: ERROR: Make sure to use same sprite dimensions in one RGAS file ***"
        break
    fi

    LEVELSET_TILE_WIDTH_PX="$(cat ${LEVELSET_INPUT} |grep  \"Width\": |uniq |perl -p -e 's/[^\d]*(\d+)[^\d]*/$1/g')"
    LEVELSET_TILE_WIDTH_PX="$(cat ${LEVELSET_INPUT} |grep  \"Height\": |uniq |perl -p -e 's/[^\d]*(\d+)[^\d]*/$1/g')"

    # FIXME: change base name or just use levelset name instead
    LEVELSET_OUTPUT_BASENAME="${LEVELSET_NAME}"

    LEVELSET_OUTPUT_H="${FOLDER_ASSETS_OUTPUT}/../../levelset_${LEVELSET_OUTPUT_BASENAME}.h"
    LEVELSET_OUTPUT_C="${FOLDER_ASSETS_OUTPUT}/../../levelset_${LEVELSET_OUTPUT_BASENAME}.c"  
    LEVELSET_OUTPUT_TEMP_C="${FOLDER_ASSETS_OUTPUT}/${LEVELSET_OUTPUT_BASENAME}.c"
    
    LEVELSET_OUTPUT="${FOLDER_ASSETS_OUTPUT}/${LEVELSET_OUTPUT_BASENAME}.levelset_map"

    LEVELSET_OUTPUT_PREFIX="${FOLDER_ASSETS_OUTPUT}/${LEVELSET_OUTPUT_BASENAME}_level_"

    # ---

    # 1) create .c file
    # #################

    ${CMD_MONO} ./rgas/RGAS.1.2.5.exe ${LEVELSET_INPUT} ${LEVELSET_OUTPUT_TEMP_C}

    perl -i -p -e 's/\r\n|\n|\r/\n/g' ${LEVELSET_OUTPUT_TEMP_C}

    # create levels file
    LEVELS_SEPARATOR="const unsigned char G_level"

    LEVELSET_FILE_HEX="$(basename $(createTempFile))"
    LEVELSET_FILE_INCLUDES="$(basename $(createTempFile))"

    rm -rf ${LEVELSET_FILE_HEX} >/dev/null 2>&1
    touch ${LEVELSET_FILE_HEX}

    #rm -rf ${LEVELSET_OUTPUT_ENUMS} >/dev/null 2>&1
    #touch ${LEVELSET_OUTPUT_ENUMS}

    rm -rf ${LEVELSET_OUTPUT} >/dev/null 2>&1
    touch ${LEVELSET_OUTPUT}

    #printf "enum {\n" >> ${LEVELSET_OUTPUT_ENUMS}
        
    LEVELSET_MAP_NAME="g_levelset_${LEVELSET_OUTPUT_BASENAME}_map"
    LEVELSET_FILL_NAME="levelset_${LEVELSET_OUTPUT_BASENAME}_map_fill"

    printf "TEOPL_MEMORY_POINTER_TYPE ${LEVELSET_MAP_NAME}[];\n\n" >> ${LEVELSET_OUTPUT}

    printf "void ${LEVELSET_FILL_NAME}() {\n" >> ${LEVELSET_OUTPUT}

    NUM_LEVELS=0
    ${CMD_CSPLIT} --digits=3 --prefix=${LEVELSET_OUTPUT_PREFIX} ${LEVELSET_OUTPUT_TEMP_C} "/${LEVELS_SEPARATOR}/+1" "{*}" >/dev/null 2>&1
    rm -rf "${LEVELSET_OUTPUT_PREFIX}000" ${LEVELSET_OUTPUT_TEMP_C}

    # level width / height
    LEVEL_WIDTHS=($(cat ${LEVELSET_INPUT} |grep \"_width\": |${CMD_CUT} -d: -f2 |${CMD_SED} 's/[ ,\n\r]//g'))
    LEVEL_HEIGHTS=($(cat ${LEVELSET_INPUT} |grep \"_height\": |${CMD_CUT} -d: -f2 |${CMD_SED} 's/[ ,\n\r]//g'))

    LEVEL_FILES=$(find ${FOLDER_ASSETS_OUTPUT} -type f -iname "${LEVELSET_OUTPUT_BASENAME}_level_*" |sort)

    rm -rf ${LEVELSET_FILE_INCLUDES} >/dev/null 2>&1
    touch ${LEVELSET_FILE_INCLUDES}    

    for LEVEL_FILE in ${LEVEL_FILES}; do

        printf "#include \"assets/binary_compressed_wincpctelera/common/level_${LEVELSET_NAME}_${NUM_LEVELS}.h\"\n" >> ${LEVELSET_FILE_INCLUDES}

        LEVELSET_FILE_BINARY="${FOLDER_ASSETS_OUTPUT}/../binary/level_${LEVELSET_NAME}_${NUM_LEVELS}.bin"

        perl -i -p -e "s/^.*G_level.*//g" ${LEVEL_FILE}
        perl -i -p -e "s/\};/,/g" ${LEVEL_FILE}

        # max full dimensions (e.g. 20x20 with frame)
        LEVEL_WIDTH=$(bc <<< "${LEVEL_WIDTHS[${NUM_LEVELS}]}")
        LEVEL_HEIGHT=$(bc <<< "${LEVEL_HEIGHTS[${NUM_LEVELS}]}")

        # inner dimensions (e.g. 20x12 -> 18x10 without frame)
        LEVEL_WIDTH_INNER=$(bc <<< "${LEVEL_WIDTH} - 2")
        LEVEL_HEIGHT_INNER=$(bc <<< "${LEVEL_HEIGHT} - 2")

        perl -i -p -e "s/0x//g" ${LEVEL_FILE}
        perl -i -p -e "s/[^0-9A-F]*//g" ${LEVEL_FILE}
 
        # remove level frame tiles (assume they are filled elsewhere)
        WT2=$(bc <<< ${LEVEL_WIDTH}*2) && WT2M4=$(bc <<< ${WT2}-4) && perl -i -p -e "s/^.{$WT2}(.*).{$WT2}$/\1/g" ${LEVEL_FILE} && perl -i -p -e "s/..(.{$WT2M4})../\1/g" ${LEVEL_FILE}

        # encode and append level dimensions (example: W=18 H=10 => (W-3)*16 + (H-3) = 15*16 + 7 = 247 = 0b 1111 0111)
        # dimensions were decreased by 3 (minimal meanigful dimension)
        LEVEL_WIDTH_INNER_ENCODED=$(bc <<< "${LEVEL_WIDTH_INNER} - 3")
        LEVEL_HEIGHT_INNER_ENCODED=$(bc <<< "${LEVEL_HEIGHT_INNER} - 3")
        INNER_DIMENSION_ENCODED=$(bc <<< "${LEVEL_WIDTH_INNER_ENCODED}*16 + ${LEVEL_HEIGHT_INNER_ENCODED}")
        INNER_DIMENSION_ENCODED_HEX=$(echo "obase=16;ibase=10;${INNER_DIMENSION_ENCODED}" |bc)
        ${CMD_SED} -i "1s/^/${INNER_DIMENSION_ENCODED_HEX}/" ${LEVEL_FILE}

        cat ${LEVEL_FILE} >> ${LEVELSET_FILE_HEX}
        rm -rf ${LEVEL_FILE} >/dev/null 2>&1

        # HEX -> BINARY
        cat ${LEVELSET_FILE_HEX} |xxd -r -p >${LEVELSET_FILE_BINARY}
        rm -rf ${LEVELSET_FILE_HEX} >/dev/null 2>&1

        echo "*** LVL: generated '${LEVELSET_FILE_BINARY}'"

        LEVELS_SIZE=$(printf '%.0f\n' $(bc <<< "${LEVEL_WIDTH_INNER} * ${LEVEL_HEIGHT_INNER} + 1")) # +1 for width/height encoded

        ZIP_SIZE=$(printf "LEVEL_${LEVELSET_NAME}_${NUM_LEVELS}_size_z" |tr '[A-Z]' '[a-z]')
        ZIP_END=$(printf "LEVEL_${LEVELSET_NAME}_${NUM_LEVELS}_end" |tr '[A-Z]' '[a-z]')
        UNZIP_SIZE="${LEVELS_SIZE}"

        printf "    ${LEVELSET_MAP_NAME}[${NUM_LEVELS} * 3 + 0] = (TEOPL_MEMORY_POINTER_TYPE)${ZIP_SIZE};\n" >>${LEVELSET_OUTPUT}
        printf "    ${LEVELSET_MAP_NAME}[${NUM_LEVELS} * 3 + 1] = (TEOPL_MEMORY_POINTER_TYPE)${ZIP_END};\n" >>${LEVELSET_OUTPUT}
        printf "    ${LEVELSET_MAP_NAME}[${NUM_LEVELS} * 3 + 2] = (TEOPL_MEMORY_POINTER_TYPE)${UNZIP_SIZE};\n" >>${LEVELSET_OUTPUT}

        # ---

        NUM_LEVELS=$((NUM_LEVELS+1))

    done

    # add size for levelset map array
    perl -i -p -e "s/${LEVELSET_MAP_NAME}\[\];/${LEVELSET_MAP_NAME}\[${NUM_LEVELS} * 3\];/g" ${LEVELSET_OUTPUT}

    # MAP (finish)
    printf "}" >> ${LEVELSET_OUTPUT}
    perl -i -p0e 's/,\n\}/\n\};\n/smg' ${LEVELSET_OUTPUT} # remove extra comma

    # ---

    # add MAP
    cat ${LEVELSET_OUTPUT} > ${LEVELSET_OUTPUT_C}
    rm -rf ${LEVELSET_OUTPUT}

    # use u8 type
    perl -i -p -e 's/const unsigned char/const u8/g' ${LEVELSET_OUTPUT_C}

    # include headers
    ${CMD_SED} -i "1s/^/#include \"levelset_${LEVELSET_OUTPUT_BASENAME}.h\"\n\n/" ${LEVELSET_OUTPUT_C}
    
    # include info
    ${CMD_SED} -i "1s/^/\/\/ REMARK: !!! DO NOT EDIT !!! This file is auto-generated.\n\n/" ${LEVELSET_OUTPUT_C}

    # fix strange leading space macos issue
    perl -i -p -e "s/\[\s*(\d+)\s*\]/\[\1\]/g" ${LEVELSET_OUTPUT_C}

    echo "*** LVL: generated '${LEVELSET_OUTPUT_C}'"

    # 2) create .h file
    # #################

    rm -rf ${LEVELSET_OUTPUT_H} >/dev/null 2>&1
    touch ${LEVELSET_OUTPUT_H}
    
    # num levels
    printf "#define " >>${LEVELSET_OUTPUT_H}
    printf "TEOPL_LEVELSET_${LEVELSET_NAME}_NUM_LEVELS ${NUM_LEVELS}\n\n" |tr '[a-z]' '[A-Z]' >>${LEVELSET_OUTPUT_H}
 
    # add includes    
    printf "#ifndef WINCPCTELERA\n\n" >> ${LEVELSET_OUTPUT_H}
    cat ${LEVELSET_FILE_INCLUDES} >> ${LEVELSET_OUTPUT_H}
    perl -i -p -e "s/binary_compressed_wincpctelera\/common/binary_compressed/g" ${LEVELSET_OUTPUT_H}
    printf "\n#else\n\n" >> ${LEVELSET_OUTPUT_H}
    cat ${LEVELSET_FILE_INCLUDES} >> ${LEVELSET_OUTPUT_H}
    echo >> ${LEVELSET_OUTPUT_H}
    printf "#endif\n\n" >> ${LEVELSET_OUTPUT_H}
    rm -rf ${LEVELSET_FILE_INCLUDES} >/dev/null 2>&1

    # variables
    cat ${LEVELSET_OUTPUT_C} |grep '^TEOPL_MEMORY_POINTER_TYPE' |${CMD_CUT} -d'[' -f1 >>${LEVELSET_OUTPUT_H}
    perl -i -p -e 's/^TEOPL_MEMORY_POINTER_TYPE(.*)/extern TEOPL_MEMORY_POINTER_TYPE$1\[\];/g' ${LEVELSET_OUTPUT_H}

    # functions
    printf "\nvoid levelset_${LEVELSET_NAME}_map_fill();\n" >>${LEVELSET_OUTPUT_H}

    ${CMD_SED} -i "1s/^/#include \"teopl_memory.h\"\n\n/" ${LEVELSET_OUTPUT_H}
    ${CMD_SED} -i "1s/^/#include <types.h>\n\n/" ${LEVELSET_OUTPUT_H}
    ${CMD_SED} -i "1s/^/#include <cpctelera.h>\n\n/" ${LEVELSET_OUTPUT_H}
    ${CMD_SED} -i "1s/^/#define _TEOPL_LEVELSET_${LEVELSET_NAME}_\n\n/" ${LEVELSET_OUTPUT_H}
    ${CMD_SED} -i "1s/^/#ifndef _TEOPL_LEVELSET_${LEVELSET_NAME}_\n\n/" ${LEVELSET_OUTPUT_H}
    ${CMD_SED} -i "1s/^/\/\/ REMARK: !!! DO NOT EDIT !!! This file is auto-generated.\n\n/" ${LEVELSET_OUTPUT_H}

    printf "\n#endif\n" >>${LEVELSET_OUTPUT_H}

    echo "*** LVL: generated '${LEVELSET_OUTPUT_H}'"
done

# error?
if [[ "$ERROR" != 'no' ]]; then
    echo "${COLOR_RED}${ERROR}${COLOR_NORMAL}"
    rm -rf ${LEVELSET_FILE_BINARY} ${LEVELSET_OUTPUT_C} ${LEVELSET_OUTPUT_H} >/dev/null 2>&1
    exit 1
fi

rm foo >/dev/null 2>&1
echo "${COLOR_GREEN}*** LVL: level server updated${COLOR_NORMAL}"

# output file
echo "${COLOR_LIGHT_GREEN}*** LVL: REGENERATED!${COLOR_NORMAL}"

echo
