#!/bin/bash

#-----------------------------LICENSE NOTICE-----------------------------------
#     __    __  ______  __________
#    / /   / / / / __ \/  _/ ____/
#   / /   / / / / / / // // /
#  / /___/ /_/ / /_/ // // /___
# /_____/\____/_____/___/\____/
# (break the loop)
#
# This file is part of Amstrad CPC game "LUDIC break the loop".
#
#  "LUDIC break the loop" is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  "LUDIC break the loop" is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http:#www.gnu.org/licenses/>.
#
#  Code, Concept, Story Copyright (C) 2019 Naad (@teopl)
#  Art, Music, Story Copyright (C) 2019 Branho (@brando464)
#
#  Contact: teopl.the.game@gmail.com
#
# Developed for the CPCRetrodev contest 2019 edition.
#
#------------------------------------------------------------------------------

# include
source $(dirname $0)/bash_library.sh

if (( $# != 2 )); then
    echo "${COLOR_RED}*** GFX: Illegal number of parameters!${COLOR_NORMAL}"
    exit 1
fi

max()
{
    local m="$1"
    for n in "$@"
    do
        [ "$n" -gt "$m" ] && m="$n"
    done
    echo "$m"
}

# #######################################################################################################################################

# ---

FOLDER_ASSETS_INPUT=$1
FOLDER_ASSETS_OUTPUT=$2

 # bail?
 if [ -d "$FOLDER_ASSETS_OUTPUT" ]; then
    printf "*** GFX: No need to regenerate...\n\n"
    exit 0
fi

mkdir -p $FOLDER_ASSETS_OUTPUT

ERROR="no"

MAX_NUM_BYTES_FOR_DECOMPRESSED_ANYTHING=0

for f in $(find ${FOLDER_ASSETS_INPUT} -type f -iname '*.rgas'); do

    FILENAME=$(basename -- "$f")
    #EXTENSION="${filename##*.}"
    SPRITE_NAME="${FILENAME%.*}"

    SPRITE_INPUT="${FOLDER_ASSETS_INPUT}/${SPRITE_NAME}.rgas"

    echo "*** GFX: check '${SPRITE_INPUT}'..."

    HAS_LEVELS="$(grep '_aspectRatio' -i ${SPRITE_INPUT} |tr -d '\0')"

    if [ "$HAS_LEVELS" != "" ]; then
        ERROR="*** GFX: ERROR: Levels are not allowed in sprite RGAS file ***"
        break
    fi

    # ---

    # check input file format
    INPUT_FORMAT1_OK="$(cat ${SPRITE_INPUT} |grep '"RGAS.PlatformProviders.AmstradSpriteSet+AmstradCPCRSLibCArrayFormatter,')"
    INPUT_FORMAT2_OK="$(cat ${SPRITE_INPUT} |grep '"OutputSpriteDimensions": true')"
    INPUT_FORMAT3_OK="$(cat ${SPRITE_INPUT} |grep  \"Width\": |uniq  |${CMD_WC} -l)$(cat ${SPRITE_INPUT} |grep \"Height\": |uniq  |${CMD_WC} -l)"
    INPUT_FORMAT3_OK=$(echo $INPUT_FORMAT3_OK | tr -d ' ')

    if [ "$INPUT_FORMAT1_OK" == "" ] || [ "$INPUT_FORMAT2_OK" == "" ]; then
        ERROR="*** GFX: ERROR: Make sure to select 'CPCTelera/SDCC C array' and 'Output dimensions for sprite' in RGAS ***"
        break
    fi
    if [ "$INPUT_FORMAT3_OK" != "11" ]; then
        ERROR="*** GFX: ERROR: Make sure to use same sprite dimensions in one RGAS file ***"
        break
    fi

    MODE=$(cat ${SPRITE_INPUT} |grep '"Mode":' |${CMD_CUT} -d: -f2 |${CMD_CUT} -d, -f1 |${CMD_CUT} -d' ' -f2)

    SPRITE_WIDTH_PX="$(cat ${SPRITE_INPUT} |grep  \"Width\": |uniq |perl -p -e 's/[^\d]*(\d+)[^\d]*/$1/g')"
    SPRITE_HEIGHT_PX="$(cat ${SPRITE_INPUT} |grep  \"Height\": |uniq |perl -p -e 's/[^\d]*(\d+)[^\d]*/$1/g')"

    #SPRITE_OUTPUT_BASENAME="mode_${MODE}_${SPRITE_NAME}"
    SPRITE_OUTPUT_BASENAME="${SPRITE_NAME}"

    SPRITE_OUTPUT_H="${FOLDER_ASSETS_OUTPUT}/${SPRITE_OUTPUT_BASENAME}.h"
    SPRITE_OUTPUT_C="${FOLDER_ASSETS_OUTPUT}/${SPRITE_OUTPUT_BASENAME}.c"

    SPRITE_OUTPUT_ENUMS="${FOLDER_ASSETS_OUTPUT}/${SPRITE_OUTPUT_BASENAME}.sprites_enums"
    SPRITE_OUTPUT_MAP="${FOLDER_ASSETS_OUTPUT}/${SPRITE_OUTPUT_BASENAME}.sprites_map"
    SPRITE_OUTPUT_PALETTE="${FOLDER_ASSETS_OUTPUT}/${SPRITE_OUTPUT_BASENAME}.sprites_palette"

    # ---

    # 1) create .c file
    # #################

    # ---

    ${CMD_MONO} ./rgas/RGAS.1.2.5.exe ${SPRITE_INPUT} ${SPRITE_OUTPUT_C}

    perl -i -p -e 's/\r\n|\n|\r/\n/g' ${SPRITE_OUTPUT_C}

    # create sprite enum file
    cat ${SPRITE_OUTPUT_C} |grep 'G_' |perl -p -e 's/.*(G_[^\[]+).*/\1/' >${SPRITE_OUTPUT_ENUMS}
    NUM_SPRITES=$(${CMD_WC} -l <${SPRITE_OUTPUT_ENUMS})

    # create sprite map file
    rm -rf ${SPRITE_OUTPUT_MAP} >/dev/null 2>&1
    touch ${SPRITE_OUTPUT_MAP}

    # create palette file
    PALETTE_PREFIX="g_$(echo ${SPRITE_OUTPUT_BASENAME})_palette_fw"
    PALETTE_ENCODED="$(perl -ne 'if (/"Inks":/../\}/) {print unless /"Inks":/ or /\}/}' ${SPRITE_INPUT} |grep value |${CMD_CUT} -d: -f2 |${CMD_CUT} -d\" -f2)"
    PALETTE_DECODED_BINARY="${SPRITE_OUTPUT_BASENAME}.palette_decoded_binary"
    ${CMD_BASE64} -d <<< ${PALETTE_ENCODED} >${PALETTE_DECODED_BINARY} && cpct_bin2c ${PALETTE_DECODED_BINARY} -an ${PALETTE_PREFIX} >${SPRITE_OUTPUT_PALETTE}
    rm -rf ${PALETTE_DECODED_BINARY} >/dev/null 2>&1

    # fix rgas output
    perl -i -p -e 's/\[(\d+)\]\s*= \{\s*\n/\[\($1 - 2\) \/ 1 \+ 2 \/\* \(w_px, h_px\) \*\/ \] = \{\n/g' ${SPRITE_OUTPUT_C}
    PIXELS_PER_BYTE=0
    case "$MODE" in
        0)
            PIXELS_PER_BYTE=2
            perl -i -p0e 's/= \{\n\s*(0x..),\s*(0x..),\s*\n/= { \n\t\/\* w_px = \*\/ $1 \* 2 \/\* \(PIXELS_PER_BYTE\) \*\/ \/ 1, \n\t\/\* h_px = \*\/ $2, \n/smg' ${SPRITE_OUTPUT_C}
            ;;
        1)
            PIXELS_PER_BYTE=4
            perl -i -p0e 's/= \{\n\s*(0x..),\s*(0x..),\s*\n/= { \n\t\/\* w_px = \*\/ $1 \* 4 \/\* \(PIXELS_PER_BYTE\) \*\/ \/ 1, \n\t\/\* h_px = \*\/ $2, \n/smg' ${SPRITE_OUTPUT_C}
            ;;
        2)
            PIXELS_PER_BYTE=8
            perl -i -p0e 's/= \{\n\s*(0x..),\s*(0x..),\s*\n/= { \n\t\/\* w_px = \*\/ $1 \* 8 \/\* \(PIXELS_PER_BYTE\) \*\/ \/ 1, \n\t\/\* h_px = \*\/ $2, \n/smg' ${SPRITE_OUTPUT_C}
            ;;
        *)
            echo "${COLOR_RED}*** GFX: ERROR: unknown mode '$MODE'${COLOR_NORMAL}"
            exit 1 
    esac
    #perl -i -p -e "s/G_/G_mode${MODE}_/g" ${SPRITE_OUTPUT_C}
    perl -i -p -e 's/\}/\n\}/g' ${SPRITE_OUTPUT_C}

    # sprite size
    SPRITE_SIZE_B=$(printf '%.0f\n' $(bc <<< "${SPRITE_WIDTH_PX} * ${SPRITE_HEIGHT_PX} / ${PIXELS_PER_BYTE}"))

    # calculate decompressed size
    MAX_NUM_BYTES_FOR_DECOMPRESSED_ANYTHING_CANDIDATE=$(printf '%.0f\n' $(bc <<< "${SPRITE_WIDTH_PX} * ${SPRITE_HEIGHT_PX} * ${NUM_SPRITES} / ${PIXELS_PER_BYTE}"))
    MAX_NUM_BYTES_FOR_DECOMPRESSED_ANYTHING=$(( MAX_NUM_BYTES_FOR_DECOMPRESSED_ANYTHING_CANDIDATE > MAX_NUM_BYTES_FOR_DECOMPRESSED_ANYTHING ?\
        MAX_NUM_BYTES_FOR_DECOMPRESSED_ANYTHING_CANDIDATE : MAX_NUM_BYTES_FOR_DECOMPRESSED_ANYTHING ))

    # remove every second byte (not needed if rgas is set properly - output format?)
    #perl -i -p -e 's/0x.., (0x..)/$1/g' ${SPRITE_OUTPUT_C}
    
    # create binary file
    FILE_BINARY="${FOLDER_ASSETS_OUTPUT}/../binary/sprites_${SPRITE_OUTPUT_BASENAME}.bin"
    #FILE_BINARY_COMPRESSED="${SPRITE_OUTPUT_BASENAME}_sprites_compressed"
    
    egrep '^\s+0x..' ${SPRITE_OUTPUT_C} | tr '\n' ' ' |${CMD_SED} 's/0x//g' |${CMD_SED} 's/\t//g' |${CMD_SED} 's/\s//g' |${CMD_SED} 's/,//g' |xxd -r -p >${FILE_BINARY}

    # compress binary file
    #${CMD_WINE} ./zx7/zx7.exe -f ${FILE_BINARY} ${FILE_BINARY_COMPRESSED} >/dev/null 2>&1
    #rm -rf ${FILE_BINARY} >/dev/null 2>&1

    # convert compressed binary to C array
    #cpct_bin2c ${FILE_BINARY_COMPRESSED} -an g_${FILE_BINARY_COMPRESSED} >${SPRITE_OUTPUT_C}
    #rm -rf ${FILE_BINARY_COMPRESSED} >/dev/null 2>&1

    # convert binary to C array
    cpct_bin2c ${FILE_BINARY} -an g_${SPRITE_OUTPUT_BASENAME} >${SPRITE_OUTPUT_C}
    rm -rf ${FILE_BINARY} >/dev/null 2>&1

    # add palette
    cat ${SPRITE_OUTPUT_PALETTE} >> ${SPRITE_OUTPUT_C}
    rm -rf ${SPRITE_OUTPUT_PALETTE} >/dev/null 2>&1

    # sprite map
    printf "/*\n//const u16 g_${SPRITE_OUTPUT_BASENAME}_sprites_map[] = {\n" >> ${SPRITE_OUTPUT_MAP}
    SPRITE_SIZE_OFFSET=0
    for line in $(cat ${SPRITE_OUTPUT_ENUMS}); do
        printf "\t${line},\n" >>${SPRITE_OUTPUT_MAP}
        SPRITE_SIZE_OFFSET=$(printf '%.0f\n' $(bc <<< "${SPRITE_SIZE_OFFSET} + ${SPRITE_SIZE_B}"))
    done
    printf "}\n*/\n" >> ${SPRITE_OUTPUT_MAP}

    perl -i -p0e 's/,\n\}/\n\};\n/smg' ${SPRITE_OUTPUT_MAP} # remove extra comma

    if [ -e "${SPRITE_OUTPUT_MAP}" ]; then
        cat ${SPRITE_OUTPUT_MAP} >>${SPRITE_OUTPUT_C}
        rm -rf ${SPRITE_OUTPUT_MAP}
    fi

    # use u8 type
    perl -i -p -e 's/const unsigned char/const u8/g' ${SPRITE_OUTPUT_C}

    # include headers
    ${CMD_SED} -i "1s/^/#include \"${SPRITE_OUTPUT_BASENAME}.h\"\n\n/" ${SPRITE_OUTPUT_C}
    
    # include info
    ${CMD_SED} -i "1s/^/\/\/ REMARK: !!! DO NOT EDIT !!! This file is auto-generated.\n\n/" ${SPRITE_OUTPUT_C}

    # fix strange leading space macos issue
    perl -i -p -e "s/\[\s*(\d+)\s*\]/\[\1\]/g" ${SPRITE_OUTPUT_C}

    echo "*** GFX: generated '${SPRITE_OUTPUT_C}'"

    # 2) create .h file
    # #################

    rm -rf ${SPRITE_OUTPUT_H} >/dev/null 2>&1
    touch ${SPRITE_OUTPUT_H}
   
    # dimensions
    SPRITE_WIDTH_B=$(bc <<< "${SPRITE_WIDTH_PX} / ${PIXELS_PER_BYTE}")
    printf "#define ${SPRITE_OUTPUT_BASENAME}_WIDTH_B ${SPRITE_WIDTH_B}\n" >>${SPRITE_OUTPUT_H}
    printf "#define ${SPRITE_OUTPUT_BASENAME}_HEIGHT_PX ${SPRITE_HEIGHT_PX}\n\n" >>${SPRITE_OUTPUT_H}

    printf "#define ${SPRITE_OUTPUT_BASENAME}_NUM_SPRITES ${NUM_SPRITES}\n" >>${SPRITE_OUTPUT_H}

    BYTES_PER_SPRITE=$(bc <<< "${SPRITE_WIDTH_PX} * ${SPRITE_HEIGHT_PX} / ${PIXELS_PER_BYTE}")
    printf "#define ${SPRITE_OUTPUT_BASENAME}_BYTES_PER_SPRITE ${BYTES_PER_SPRITE}\n\n" >>${SPRITE_OUTPUT_H}

    # sprite enums
    printf "/*\nenum {" >>${SPRITE_OUTPUT_H}
    SPRITE_SIZE_OFFSET=0
    for line in $(cat ${SPRITE_OUTPUT_ENUMS}); do
        printf "\t${line} = ${SPRITE_SIZE_OFFSET},\n" >>${SPRITE_OUTPUT_H}
        SPRITE_SIZE_OFFSET=$(printf '%.0f\n' $(bc <<< "${SPRITE_SIZE_OFFSET} + ${SPRITE_SIZE_B}"))
    done
    printf "\n\t${SPRITE_OUTPUT_BASENAME}_SPRITES_NUM_BYTES = ${SPRITE_SIZE_OFFSET}\n};\n*/\n" |tr '[a-z]' '[A-Z]' >> ${SPRITE_OUTPUT_H}
    rm -rf ${SPRITE_OUTPUT_ENUMS}

    # variable
    cat ${SPRITE_OUTPUT_C} |grep 'const u8' |${CMD_CUT} -d'[' -f1 >>${SPRITE_OUTPUT_H}
    cat ${SPRITE_OUTPUT_C} |grep 'const u16' |${CMD_CUT} -d'[' -f1 >>${SPRITE_OUTPUT_H}
    perl -i -p -e 's/const u8(.*)/extern const u8$1\[\];/g' ${SPRITE_OUTPUT_H}
    perl -i -p -e 's/const u16(.*)/extern const u16$1\[\];/g' ${SPRITE_OUTPUT_H}

    ${CMD_SED} -i "1s/^/#include <types.h>\n\n/" ${SPRITE_OUTPUT_H}
    ${CMD_SED} -i "1s/^/#define _TEOPL_${SPRITE_NAME}_\n\n/" ${SPRITE_OUTPUT_H}
    ${CMD_SED} -i "1s/^/#ifndef _TEOPL_${SPRITE_NAME}_\n\n/" ${SPRITE_OUTPUT_H}
    ${CMD_SED} -i "1s/^/\/\/ REMARK: !!! DO NOT EDIT !!! This file is auto-generated.\n\n/" ${SPRITE_OUTPUT_H}

    printf "\n#endif\n" >>${SPRITE_OUTPUT_H}

    echo "*** GFX: generated '${SPRITE_OUTPUT_H}'"
done

# error?
if [[ "$ERROR" != 'no' ]]; then
    echo "${COLOR_RED}${ERROR}${COLOR_NORMAL}"
    rm -rf ${SPRITE_OUTPUT_C} ${SPRITE_OUTPUT_H} >/dev/null 2>&1
    exit 1
fi

# output file
echo "${COLOR_LIGHT_GREEN}*** GFX: REGENERATED!${COLOR_NORMAL}"

echo
