#!/bin/bash

#-----------------------------LICENSE NOTICE-----------------------------------
#     __    __  ______  __________
#    / /   / / / / __ \/  _/ ____/
#   / /   / / / / / / // // /
#  / /___/ /_/ / /_/ // // /___
# /_____/\____/_____/___/\____/
# (break the loop)
#
# This file is part of Amstrad CPC game "LUDIC break the loop".
#
#  "LUDIC break the loop" is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  "LUDIC break the loop" is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http:#www.gnu.org/licenses/>.
#
#  Code, Concept, Story Copyright (C) 2019 Naad (@teopl)
#  Art, Music, Story Copyright (C) 2019 Branho (@brando464)
#
#  Contact: teopl.the.game@gmail.com
#
# Developed for the CPCRetrodev contest 2019 edition.
#
#------------------------------------------------------------------------------

# include
source $(dirname $0)/bash_library.sh

if (( $# != 2 )); then
    echo "${COLOR_RED}*** TXT: Illegal number of parameters!${COLOR_NORMAL}"
    exit 1
fi

max()
{
    local m="$1"
    for n in "$@"
    do
        [ "$n" -gt "$m" ] && m="$n"
    done
    echo "$m"
}

# ---

FOLDER_ASSETS_INPUT=$1
FOLDER_ASSETS_OUTPUT=$2

 # bail?
 if [ -d "$FOLDER_ASSETS_OUTPUT" ]; then
    printf "*** TXT: No need to regenerate...\n\n"
    exit 0
fi

ERROR="no"

SIZE1=0
SIZE2=0

for f in $(find ${FOLDER_ASSETS_INPUT} -type f -iname '*.txt'); do

    FILENAME=$(basename -- "$f")
    #EXTENSION="${filename##*.}"
    TEXT_NAME="${FILENAME%.*}"

    TEXT_INPUT="${FOLDER_ASSETS_INPUT}/${TEXT_NAME}.txt"

    perl -i -p -e 's/\r\n|\n|\r/\n/g' ${TEXT_INPUT}

    echo "*** TXT: check '${TEXT_INPUT}'..."

    # ---

    TEXT_OUTPUT_BASENAME="${TEXT_NAME}"

    TEXT_OUTPUT_H="${FOLDER_ASSETS_OUTPUT}/${TEXT_OUTPUT_BASENAME}.h"
    TEXT_OUTPUT_C="${FOLDER_ASSETS_OUTPUT}/${TEXT_OUTPUT_BASENAME}.c"

    WORD_USAGE="${FOLDER_ASSETS_OUTPUT}/word_usage"
    WORD_USAGE_C="${FOLDER_ASSETS_OUTPUT}/word_usage.c"
    WORD_USAGE_H="${FOLDER_ASSETS_OUTPUT}/word_usage.h"
    
    mkdir -p $FOLDER_ASSETS_OUTPUT >/dev/null 2>&1

    if [ ! -f "${WORD_USAGE}" ]; then

        printf "*** TXT: Generate '${WORD_USAGE}'..."

        rm -rf foo >/dev/null 2>&1
        cat ${TEXT_INPUT} |grep -v '^#' |perl -p -e 's/^........//g' |perl -p -e "s/[^0-9A-Z \n\r']+/ /g" | tr ' ' '\n' | sort | uniq -c |tail --lines=+2 |sort >foo

        rm -rf ${WORD_USAGE} >/dev/null 2>&1
        while IFS1='' read -r LINE || [[ -n "$line" ]]; do

            WORD=$(echo ${LINE} |cut -d' ' -f2)    

            REPEATED=$(echo ${LINE} |cut -d' ' -f1)

            LENGTH_MINUS_ONE=${#WORD}
            LENGTH_MINUS_ONE=$(bc <<< "${LENGTH_MINUS_ONE} - 1")

            SAVED_BYTES=$(bc <<< "${REPEATED} * ${LENGTH_MINUS_ONE}")

            #printf %05s "${SAVED_BYTES}"
            #printf " \"${WORD}\" ${REPEATED} * ${LENGTH_MINUS_ONE}\n"

            printf %05s "${SAVED_BYTES}" >>${WORD_USAGE}
            printf " \"${WORD}\" ${REPEATED} * ${LENGTH_MINUS_ONE}\n" >>${WORD_USAGE}

        done < foo

        rm -rf foo >/dev/null 2>&1

        printf " done!\n"
    fi

    if [ ! -f "${WORD_USAGE_C}" ]; then

        # 0) generate WORD USAGE
        # ###

        printf "*** TXT: Generate '${WORD_USAGE_H}'..."

        # create WORD USAGE H (for decoding)
        printf "extern char* const frequent_words[];\n\n" >${WORD_USAGE_H}
        printf "#endif\n" >>${WORD_USAGE_H}
        ${CMD_SED} -i "1s/^/#define _WORD_USAGE_\n\n/" ${WORD_USAGE_H}
        ${CMD_SED} -i "1s/^/#ifndef _WORD_USAGE_\n\n/" ${WORD_USAGE_H}
        ${CMD_SED} -i "1s/^/\/\/ REMARK: !!! DO NOT EDIT !!! This file is auto-generated.\n\n/" ${WORD_USAGE_H}

        printf " done!\n"

        # ###

        printf "*** TXT: Generate '${WORD_USAGE_C}'..."

        # REMARK: 48 + 1(' ') + 1('_') + 204 + 1(...) = 255

        # create WORD USAGE C (for decoding)
        cat ${WORD_USAGE} |sort -r |head -n204 >foo

        WORDS_C="$(cat foo |perl -p -e 's/.*(\"[^\"]*\").*/\1,/g' |perl -p -e 's/,/\n/g' |perl -p -e 's/^[\s]+//g' |awk '{ print length, $0 }' |sort -n -s -r |cut -d" " -f2-)"
        printf "char * const frequent_words[] = {\n" >${WORD_USAGE_C}
        printf "${WORDS_C}\n" >>${WORD_USAGE_C}
        perl -i -p -e 's/"\n/",\n/g' ${WORD_USAGE_C}

        # REMARK: added special words like "..." since it's used often
        printf "\n\"...\"" >>${WORD_USAGE_C}
        printf "\n};" >>${WORD_USAGE_C}

        rm -rf foo >/dev/null 2>&1

        # create WORD USAGE BASH ARRAY HEX (for encoding)
        WORDS_BASH=$(echo ${WORDS_C} |perl -p -e 's/"//g')
        declare -a WORDS_BASH_ARRAY=(${WORDS_BASH})
        declare -a WORDS_BASH_ARRAY_HEX
        for WORD in "${WORDS_BASH_ARRAY[@]}"
        do
            WORD_HEX=$(echo -e "${WORD}\c" |xxd -p)
            WORDS_BASH_ARRAY_HEX+=(${WORD_HEX})

            #printf "'${WORD_HEX}'\n"
        done

        printf " done!\n"

        # ###

        # 1) create .c file
        # #################

        # create binary file
        #FILE_BINARY="${TEXT_OUTPUT_BASENAME}.file_binary"
        FILE_BINARY="${FOLDER_ASSETS_OUTPUT}/../binary/texts_${TEXT_OUTPUT_BASENAME}.bin"
        #FILE_BINARY_COMPRESSED="${TEXT_OUTPUT_BASENAME}_compressed"

        printf "*** TXT: Generate '${FILE_BINARY}'..."

        rm -rf ${FILE_BINARY} >/dev/null 2>&1
        touch ${FILE_BINARY}

        SIZE_ENCODED_LINE_MAX=0
        while IFS1='' read -r STRING_LINE || [[ -n "$line" ]]; do

            COMMENT=$(echo $STRING_LINE |grep '^#')
        
            if [ "$COMMENT" != "" ]; then
                continue
            fi

            STRING_LINE=$(echo ${STRING_LINE} |perl -p -e 's/ +/ /g')
            STRING_LINE=$(echo ${STRING_LINE} |perl -p -e 's/ _/_/g')
            STRING_LINE=$(echo ${STRING_LINE} |perl -p -e 's/_ /_/g')

            # 1) texts start with size
            # 2) then 'B', 'C', 'D', 'E' depending on type:
            #   - welcome message AI    = 'B'
            #   - welcome message Prof  = 'C'
            #   - level message AI      = 'D'
            #   - level message Prof    = 'E'

            STRING_LEVEL_INDEX=$(echo ${STRING_LINE} |perl -p -e 's/....0*(\d+).*/\1/g')
            STRING_LEVEL_INDEX_HEX=$(printf "%x" ${STRING_LEVEL_INDEX}) # includes terminating character
 
            TEXT=$(echo "${STRING_LINE}" |perl -p -e 's/......: (.*)/\1/g')

            TEXT_TYPE="?"

            if [[ "$(echo ${STRING_LINE} |egrep '^WLCA')" != "" ]]; then
                TEXT_TYPE="B"
            elif [[ "$(echo ${STRING_LINE} |egrep '^WLCP')" != "" ]]; then
                TEXT_TYPE="C"
            elif [[ "$(echo ${STRING_LINE} |egrep '^LVLA')" != "" ]]; then
                TEXT_TYPE="D"
            elif [[ "$(echo ${STRING_LINE} |egrep '^LVLP')" != "" ]]; then
                TEXT_TYPE="E"
            fi

            if [[ "$TEXT_TYPE" == "?" ]]; then
                ERROR="*** TXT: ERROR: Make sure that text type is correct ***"
                break
            fi

            TEXT_SIZE_UNZIP=${#TEXT}
            TEXT_SIZE_UNZIP_HEX=$(printf "%x" ${TEXT_SIZE_UNZIP}) # includes terminating character

            #printf "STRING_LINE='${STRING_LINE}'\n"
       
            if [ "${TEXT_SIZE_UNZIP}" -gt 107 ]; then
                printf "${TEXT_SIZE_UNZIP}\n"
                ERROR="*** TXT: ERROR: Make sure that text is not longer then 107 characters in text ***"
                break
            fi

            # convert ascii character to font sprite index
            # ###

            rm -rf foo >/dev/null 2>&1
            touch foo

            echo -e "${TEXT}\c" |xxd -p -c 8192 >foo
            
            # these are indexed not occupied by the ASCII characters used (every index corresponds to one encoded word)
            declare -a INDEXES_FREE=(0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 34 35 36 37 38 42 43 47 59 61 91 92 93 94 96 97 98 99 100 101 102 103 104 105 106 107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 138 139 140 141 142 143 144 145 146 147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 195 196 197 198 199 200 201 202 203 204 205 206 207 208 209 210 211 212 213 214 215 216 217 218 219 220 221 222 223 224 225 226 227 228 229 230 231 232 233 234 235 236 237 238 239 240 241 242 243 244 245 246 247 248 249 250 251 252 253 254 255)
            
            INDEX=0
            for WORD_HEX in "${WORDS_BASH_ARRAY_HEX[@]}"
            do
                FOUND=$(cat foo |grep ${WORD_HEX})
        
                if [ "${FOUND}" != "" ]; then
                    WORD_ENCODED=${INDEXES_FREE[$INDEX]}
                    WORD_ENCODED=$(printf %02x ${WORD_ENCODED})

                    perl -i -p -e "s/${WORD_HEX}/${WORD_ENCODED}/g" foo

                    #printf "${WORDS_BASH_ARRAY[$INDEX]}(${WORDS_BASH_ARRAY_HEX[$INDEX]}) => '${WORD_ENCODED}'\n"
                fi

                INDEX=$((INDEX+1))
            done

            tr -d "\n\r" <foo >foobar

            #

            rm -rf barfoo >/dev/null 2>&1
            touch barfoo

            cat foobar |xxd -r -p >barfoo

            TEXT_SIZE_ZIP=$(wc -c barfoo |cut -d' ' -f1)
            SIZE_ENCODED_LINE_MAX=$(( TEXT_SIZE_ZIP > SIZE_ENCODED_LINE_MAX ? TEXT_SIZE_ZIP : SIZE_ENCODED_LINE_MAX ))

            SIZE1=$(bc <<< "${SIZE1} + ${TEXT_SIZE_UNZIP}")
            SIZE2=$(bc <<< "${SIZE2} + ${TEXT_SIZE_ZIP}")

            TEXT_SIZE_ZIP_HEX=$(printf "%x" ${TEXT_SIZE_ZIP}) # includes terminating character

            #printf "TEXT='${TEXT}'\n"
            #printf "STRING_LEVEL_INDEX_HEX='${STRING_LEVEL_INDEX_HEX}'\n"
            #printf "TEXT_SIZE_UNZIP='${TEXT_SIZE_UNZIP}'\n"
            #printf "TEXT_SIZE_UNZIP_HEX='${TEXT_SIZE_UNZIP_HEX}'\n"
            #printf "TEXT_SIZE_ZIP='${TEXT_SIZE_ZIP}'\n"
            #printf "TEXT_SIZE_ZIP_HEX='${TEXT_SIZE_ZIP_HEX}'\n"
            #printf "TEXT_TYPE='${TEXT_TYPE}'\n\n"

            printf "%02s\n" ${STRING_LEVEL_INDEX_HEX} |xxd -r -p >>${FILE_BINARY}        
            printf "%02s\n" ${TEXT_SIZE_ZIP_HEX} |xxd -r -p >>${FILE_BINARY}
            printf "${TEXT_TYPE}" >>${FILE_BINARY}
            cat barfoo >>${FILE_BINARY}
            #echo -n ${TEXT} >>${FILE_BINARY}

        done < ${TEXT_INPUT}

        rm -rf foo bar foobar barfoo >/dev/null 2>&1
        
        # compress binary file
        #${CMD_WINE} ./zx7/zx7.exe -f ${FILE_BINARY} ${FILE_BINARY_COMPRESSED} >/dev/null 2>&1
    
        #LEN_BINARY=$(wc -c ${FILE_BINARY} |cut -d' ' -f1)
        #LEN_BINARY_COMPRESSED=$(wc -c ${FILE_BINARY_COMPRESSED} |cut -d' ' -f1)

        #echo "*** TXT: BIN SIZE (${LEN_BINARY}) => ZIP SIZE (${LEN_BINARY_COMPRESSED})"

        #rm -rf ${FILE_BINARY} >/dev/null 2>&1

        # convert compressed binary to C array
        #cpct_bin2c ${FILE_BINARY_COMPRESSED} -an g_${FILE_BINARY_COMPRESSED} >${TEXT_OUTPUT_C}

        #rm -rf ${FILE_BINARY_COMPRESSED} >/dev/null 2>&1

        # use u8 type
        #perl -i -p -e 's/const unsigned char/const u8/g' ${TEXT_OUTPUT_C}

        # include headers
        #${CMD_SED} -i "1s/^/#include \"${TEXT_OUTPUT_BASENAME}.h\"\n\n/" ${TEXT_OUTPUT_C}
    
        # include info
        #${CMD_SED} -i "1s/^/\/\/ REMARK: !!! DO NOT EDIT !!! This file is auto-generated.\n\n/" ${TEXT_OUTPUT_C}

        printf " done!\n"

        printf "\n*** TXT: RAW SIZE (${SIZE1}) => ENCODED SIZE (${SIZE2})\n"

        # 2) create .h file
        # #################

        #printf "*** TXT: Generate '${TEXT_OUTPUT_C}'..."

        #rm -rf ${TEXT_OUTPUT_H} >/dev/null 2>&1
        #touch ${TEXT_OUTPUT_H}
    
        # variable
        #cat ${TEXT_OUTPUT_C} |grep 'const u8' |${CMD_CUT} -d'[' -f1 >>${TEXT_OUTPUT_H}
        #cat ${TEXT_OUTPUT_C} |grep 'const u16' |${CMD_CUT} -d'[' -f1 >>${TEXT_OUTPUT_H}
        #perl -i -p -e 's/const u8(.*)/extern const u8$1\[\];/g' ${TEXT_OUTPUT_H}
        #perl -i -p -e 's/const u16(.*)/extern const u16$1\[\];/g' ${TEXT_OUTPUT_H}

        #${CMD_SED} -i "1s/^/#include <types.h>\n\n/" ${TEXT_OUTPUT_H}    
        #${CMD_SED} -i "1s/^/#include \"word_usage.h\"\n\n/" ${TEXT_OUTPUT_H}    
        #${CMD_SED} -i "1s/^/#define _TEOPL_${TEXT_NAME}_\n\n/" ${TEXT_OUTPUT_H}
        #${CMD_SED} -i "1s/^/#ifndef _TEOPL_${TEXT_NAME}_\n\n/" ${TEXT_OUTPUT_H}
        #${CMD_SED} -i "1s/^/\/\/ REMARK: !!! DO NOT EDIT !!! This file is auto-generated.\n\n/" ${TEXT_OUTPUT_H}

        #printf "\n#endif\n" >>${TEXT_OUTPUT_H}

        #printf " done!\n"

        #echo "*** TXT: generated '${TEXT_OUTPUT_H}'"

        fi
done

# error?
if [[ "$ERROR" != 'no' ]]; then
    echo "${COLOR_RED}${ERROR}${COLOR_NORMAL}"
    rm -rf ${TEXT_OUTPUT_C} ${TEXT_OUTPUT_H} >/dev/null 2>&1
    exit 1
fi

# write max encoded line size to save space in text buffer
echo "${COLOR_YELLOW}*** TXT: Max encoded line bytes: 107 => ${SIZE_ENCODED_LINE_MAX} bytes${COLOR_NORMAL}"
perl -i -p -e "s/TEOPL_LEVEL_TEXTS_ENCODED_LINE\s+(\d+)/TEOPL_LEVEL_TEXTS_ENCODED_LINE ${SIZE_ENCODED_LINE_MAX}/g" ../src/teopl_level.h 

# output file
echo "${COLOR_LIGHT_GREEN}*** TXT: REGENERATED!${COLOR_NORMAL}"

echo
