#!/bin/bash

#-----------------------------LICENSE NOTICE-----------------------------------
#     __    __  ______  __________
#    / /   / / / / __ \/  _/ ____/
#   / /   / / / / / / // // /
#  / /___/ /_/ / /_/ // // /___
# /_____/\____/_____/___/\____/
# (break the loop)
#
# This file is part of Amstrad CPC game "LUDIC break the loop".
#
#  "LUDIC break the loop" is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  "LUDIC break the loop" is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <http:#www.gnu.org/licenses/>.
#
#  Code, Concept, Story Copyright (C) 2019 Naad (@teopl)
#  Art, Music, Story Copyright (C) 2019 Branho (@brando464)
#
#  Contact: teopl.the.game@gmail.com
#
# Developed for the CPCRetrodev contest 2019 edition.
#
#------------------------------------------------------------------------------

# Update <LEVELS> rgas with sprites from new <SPRITE> rgas file
# 1. from   <LEVELS>        create <MAP1>                           ( <SPRITE_INDEX_OLD>    =>  <SPRITE_NAME>)
# 2. from   <SPRITES>       create <MAP2>                           ( <SPRITE_NAME>         =>  <SPRITE_INDEX_NEW>)
# 3. from   <MAP1>,<MAP2>   create <MAP3>                           ( <SPRITE_INDEX_OLD>    =>  <SPRITE_INDEX_NEW>)
# 3. from   <SPRITES>       copy sprites to <LEVELS>
# 4. from   <MAP3>          replace level indexes in <LEVELS>

# include
source $(dirname $0)/bash_library.sh

if (( $# != 2 )); then
    echo "${COLOR_RED}*** LVL_UPDATE: Illegal number of parameters!${COLOR_NORMAL}"
    exit 1
fi

if [[ "$OSTYPE" == "linux-gnu" ]]; then
	CMD_CSPLIT="csplit"
elif [[ "$OSTYPE" == "darwin"* ]]; then
	CMD_CSPLIT="gcsplit"
elif [[ "$OSTYPE" == "cygwin" ]]; then
	CMD_CSPLIT="csplit"
else
	echo "${COLOR_RED}*** LVL_UPDATE: OS ${OSTYPE} not supported!${COLOR_NORMAL}"; exit 1;
fi

# #######################################################################################################################################

new_index_of_name() { i=-1; for ((j=0;j<${#INDEX_NEW_TO_NAME[@]};j++)); do [ "${INDEX_NEW_TO_NAME[$j]}" = "$1" ] && { i=$j; break; } done; echo $i; }

# ---

FILE_SPRITES=$1
FILE_LEVELS=$2

# foo0 (level sprites part)
# foo1 (level levels part)
${CMD_CSPLIT} --digits=1 --prefix=foo ${FILE_LEVELS} "/_levelList/+1" "{*}" >/dev/null 2>&1

# bar0 (blocks sprites part)
# bar1 (blocks levels part) <- should be empty
${CMD_CSPLIT} --digits=1 --prefix=bar ${FILE_SPRITES} "/_levelList/+1" "{*}" >/dev/null 2>&1

echo "EMPTY" >foo
cat foo0 |grep _name |perl -p -e 's/.*: \"(.*)\".*/\1/g' >>foo
readarray -t INDEX_OLD_TO_NAME < foo

echo "EMPTY" >foo
cat ${FILE_SPRITES} |grep _name |perl -p -e 's/.*: \"(.*)\".*/\1/g' >>foo
readarray -t INDEX_NEW_TO_NAME < foo

rm -rf foo >/dev/null 2>&1

declare -a MAP3

INDEX_OLD=0
for i in "${INDEX_OLD_TO_NAME[@]}"
do
    INDEX_NEW=$(new_index_of_name $i)
    MAP3[$INDEX_OLD]=$INDEX_NEW
    INDEX_OLD=$((INDEX_OLD+1))
done
    
#echo "${MAP3[*]}"
#echo ${MAP3[63]}

INDEX_OLD=0
for INDEX_NEW in "${MAP3[@]}"
do
    #echo "${INDEX_OLD} to ${INDEX_NEW}"
    perl -i -p -e "s/ ([\s]+)${INDEX_OLD}([,]*[\n\r]+)/\1 ${INDEX_NEW} \2/g" foo1

    INDEX_OLD=$((INDEX_OLD+1))    
done

perl -i -p -e "s/ ,/,/g" foo1
cat bar0 foo1 >${FILE_LEVELS}

rm -rf foo0 foo1 bar0 bar1 >/dev/null 2>&1

printf "*** UPD: levelsets sprites updated: ${FILE_SPRITES} => ${FILE_LEVELS} done!\n"
