//-----------------------------LICENSE NOTICE-----------------------------------
//     __    __  ______  __________
//    / /   / / / / __ \/  _/ ____/
//   / /   / / / / / / // // /
//  / /___/ /_/ / /_/ // // /___
// /_____/\____/_____/___/\____/
// (break the loop)
//
// This file is part of Amstrad CPC game "LUDIC break the loop".
//
//  "LUDIC break the loop" is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  "LUDIC break the loop" is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//  Code, Concept, Story Copyright (C) 2019 Naad (@teopl)
//  Art, Music, Story Copyright (C) 2019 Branho (@brando464)
//
//  Contact: teopl.the.game@gmail.com
//
// Developed for the CPCRetrodev contest 2019 edition.
//
//------------------------------------------------------------------------------

#include "teopl_debug.h"

#ifdef DEBUG

#include <types.h>

#include "teopl_pool.h"
#include "teopl_level.h"
#include "teopl_video.h"
#include "teopl_font.h"
#include "teopl_interrupt.h"
#include "teopl_memory.h"
#include "teopl_sound.h"
#include "teopl_logic.h"
#include "teopl_input.h"

// ---

TeoplDebugShowType              g_teopl_debug_show;

u16                             g_teopl_debug_time_start[8];
u16                             g_teopl_debug_time_end[8];
u16                             g_teopl_debug_time_diff[8];

u8                              g_teopl_debug1_u8;
u8                              g_teopl_debug2_u8;
u16                             g_teopl_debug_u16;
u32                             g_teopl_debug_u32;

// ---

typedef struct {

    u8                          key[TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES];
    TeoplDebugValueSize         value_size;
    TEOPL_MEMORY_POINTER_TYPE   value_ptr;
    u32                         value_old;

} TeoplDebugInfo;

static TeoplDebugInfo           _info[24];

static char                     _string[TEOPL_DEBUG_INFO_LINE_NUM_BYTES];
static u8                       _num;

// ---

static void _draw_keys() {

    u8 i = 0;
    u8 y_px = TEOPL_DEBUG_PRINT_Y_PX;

    // ---

    teopl_video_draw_box_color(0x00, TEOPL_DEBUG_PRINT_X_B, TEOPL_DEBUG_PRINT_Y_PX, TEOPL_DEBUG_PRINT_W_B, font_HEIGHT_PX * _num);

    _string[TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES] = ':';
    _string[TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES + 1] = 0;

    for (i = 0; i < _num; ++i) {

        cpct_memset(_string, ' ', TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
        cpct_memcpy(_string, _info[i].key, TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);

        teopl_font_sync_print(TEOPL_DEBUG_PRINT_X_B, TEOPL_DEBUG_PRINT_Y_PX, TEOPL_DEBUG_PRINT_W_B, (char*)&_string);

        y_px += font_HEIGHT_PX;
    }
}

void teopl_debug_init() {

    g_teopl_debug_show = TEOPL_DEBUG_SHOW_TYPE_NONE;

    cpct_memset(_string, ' ', TEOPL_DEBUG_INFO_LINE_NUM_BYTES);

    g_teopl_debug1_u8 = 0;
    g_teopl_debug2_u8 = 0;
    g_teopl_debug_u16 = 0;
    g_teopl_debug_u32 = 0;

    _num = 0;

    // ---

    /*
    cpct_memcpy(_info[_num].key, "DBG1 8", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&(g_teopl_debug1_u8));
    _num++;

    cpct_memcpy(_info[_num].key, "DBG2 8", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&(g_teopl_debug2_u8));
    _num++;
    */

    /*
    cpct_memcpy(_info[_num].key, "INTRPT", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&(g_teopl_interrupt_started));
    _num++;
    */

    /*
    cpct_memcpy(_info[_num].key, "DBG 16", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_16;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&(g_teopl_debug_u16));
    _num++;

    cpct_memcpy(_info[_num].key, "DBG 32", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_32;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&(g_teopl_debug_u32));
    _num++;
    */

    /*
    cpct_memcpy(_info[_num].key, "CHP NO", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&(g_teopl_level_data.level_special_num_chips_destroyed));
    _num++;
    */

    /*
    cpct_memcpy(_info[_num].key, "SNG NO", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&(g_teopl_sound_song_index));
    _num++;

    cpct_memcpy(_info[_num].key, "SFX NO", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&(g_teopl_sound_sfx_index));
    _num++;
    */

    /*
    cpct_memcpy(_info[_num].key, "ENT NO", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&(g_teopl_level_data.entites.num_elements));
    _num++;
    */

    /*
    cpct_memcpy(_info[_num].key, "BT    ", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&g_teopl_input_bullet_time);
    _num++;
    */

    /*
    cpct_memcpy(_info[_num].key, "C1    ", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&g_teopl_video_debug_palette_c1);
    _num++;

    cpct_memcpy(_info[_num].key, "C2    ", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&g_teopl_video_debug_palette_c2);
    _num++;

    cpct_memcpy(_info[_num].key, "C3    ", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&g_teopl_video_debug_palette_c3);
    _num++;

    cpct_memcpy(_info[_num].key, "C4    ", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    _info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_8;
    _info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&g_teopl_video_debug_palette_c4);
    _num++;
    */

    //cpct_memcpy(_info[_num].key, "TIME  ", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    //_info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_32;
    //_info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&g_teopl_interrupt_ticks_total);
    //_num++;

    //cpct_memcpy(_info[_num].key, "  DIF1", TEOPL_DEBUG_INFO_LINE_KEY_NUM_BYTES);
    //_info[_num].value_size = TEOPL_DEBUG_VALUE_SIZE_16;
    //_info[_num].value_ptr = (TEOPL_MEMORY_POINTER_TYPE)(&g_teopl_debug_time_diff[0]);
    //_num++;
}

void teopl_debug_draw() {

    u8  i = 0, j = 0;
    u8  y_px = TEOPL_DEBUG_PRINT_Y_PX;
    u32 value_new;
    u8  force_show = 0;
    u8  _length;

    // ---
    
    if (g_teopl_debug_show == TEOPL_DEBUG_SHOW_TYPE_NONE) {
        return;
    }
    else if (g_teopl_debug_show == TEOPL_DEBUG_SHOW_TYPE_REFRESH) {

        _draw_keys();

        force_show = 1;
        g_teopl_debug_show = TEOPL_DEBUG_SHOW_TYPE_ON_CHANGE;
    }

    for (i = 0; i < _num; ++i) {

        if (_info[i].value_size == TEOPL_DEBUG_VALUE_SIZE_8) {
            value_new = *((u8*)(_info[i].value_ptr));
        }
        else if (_info[i].value_size == TEOPL_DEBUG_VALUE_SIZE_16) {
            value_new = *((u16*)(_info[i].value_ptr));
        }
        else if (_info[i].value_size == TEOPL_DEBUG_VALUE_SIZE_32) {
            value_new = *((u32*)(_info[i].value_ptr));
        }
#ifdef DEBUG
        else {
            TEOPL_DEBUG_PRINT("PROBLEM?");
            return;
        }
#endif

        if (force_show || _info[i].value_old != value_new) {

            // ---

            cpct_memset(_string, ' ', TEOPL_DEBUG_INFO_LINE_VALUE_NUM_BYTES);

            _info[i].value_old = value_new;

            teopl_util_u32_to_string(_string, &_length, value_new, 0);

            teopl_video_draw_box_color(0x00, TEOPL_DEBUG_INFO_LINE_VALUE_X_B, y_px, TEOPL_DEBUG_INFO_LINE_VALUE_W_B, font_HEIGHT_PX);
            
            teopl_font_sync_print(
                TEOPL_DEBUG_INFO_LINE_VALUE_X_B + font_WIDTH_B * (TEOPL_DEBUG_INFO_LINE_VALUE_NUM_BYTES - _length),
                y_px,
                TEOPL_DEBUG_INFO_LINE_VALUE_W_B,
                _string
            );
        }

        y_px += font_HEIGHT_PX;
    }
}

void teopl_debug_print(char* message) {
    teopl_font_sync_print(0, 0, TEOPL_DEFINES_VIDEO_PARAM_SCREEN_WIDTH_B, message);
}

void teopl_debug_print_u32(char* message, u32* value, TeoplDebugValueSize type, u8 num) {

    char    str_num[11];
    u8      str_num_len;
    char    buffer[255];
    char*   ptr;

    u8      i = 0;
    u32*    _value_ptr = value;
    u32     _value;

    // ---

    cpct_memset(buffer, 0, 255);

    for (ptr = buffer; *message; ++ptr, ++message) {
        *ptr = *message;
    }

    while (num--) {
        _value_ptr += i;

        if (type == TEOPL_DEBUG_VALUE_SIZE_8) {
            _value = *((u8*)_value_ptr);
        }
        else if (type == TEOPL_DEBUG_VALUE_SIZE_16) {
            _value = *((u16*)_value_ptr);
        }
        else if (type == TEOPL_DEBUG_VALUE_SIZE_32) {
            _value = *((u32*)_value_ptr);
        }

        teopl_util_u32_to_string(str_num, &str_num_len, _value, 0);

        cpct_memcpy(ptr, str_num, str_num_len);
        ptr += str_num_len;

        *ptr = ' ';
        ptr++;

        i++;
    }

    teopl_debug_print(buffer);
}

void teopl_debug_toggle() {

    if (g_teopl_debug_show == TEOPL_DEBUG_SHOW_TYPE_NONE) {
        g_teopl_debug_show = TEOPL_DEBUG_SHOW_TYPE_REFRESH;

        _draw_keys();
    }
    else {
        g_teopl_debug_show = TEOPL_DEBUG_SHOW_TYPE_NONE;

        teopl_video_draw_box_color(0x00, TEOPL_DEBUG_PRINT_X_B, TEOPL_DEBUG_PRINT_Y_PX, TEOPL_DEBUG_PRINT_W_B, font_HEIGHT_PX * _num);
    }
}

#endif // DEBUG
