//-----------------------------LICENSE NOTICE-----------------------------------
//     __    __  ______  __________
//    / /   / / / / __ \/  _/ ____/
//   / /   / / / / / / // // /
//  / /___/ /_/ / /_/ // // /___
// /_____/\____/_____/___/\____/
// (break the loop)
//
// This file is part of Amstrad CPC game "LUDIC break the loop".
//
//  "LUDIC break the loop" is free software: you can redistribute it and/or modify
//  it under the terms of the GNU Lesser General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  "LUDIC break the loop" is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//  Code, Concept, Story Copyright (C) 2019 Naad (@teopl)
//  Art, Music, Story Copyright (C) 2019 Branho (@brando464)
//
//  Contact: teopl.the.game@gmail.com
//
// Developed for the CPCRetrodev contest 2019 edition.
//
//------------------------------------------------------------------------------

#include <cpctelera.h>

#ifdef ENABLE_M4_EXPANSION

#include <types.h>

#include "teopl_tcpc.h"

#include "teopl_pool.h"
#include "teopl_sound.h"
#include "teopl_level.h"
#include "teopl_interrupt.h"
#include "teopl_debug.h"    

// ---

u8 g_teopl_tcpc_server_ip[TEOPL_TCPC_SIZE_SERVER_IP];
u8 g_teopl_tcpc_msg_send[TEOPL_TCPC_SIZE_MSG_SEND];

u8 g_teopl_tcpc_m4_rom_num;

// ---

#ifndef WINCPCTELERA

static void _send_receive(u8* msg_send, u8* msg_receive) __z88dk_callee __naked {

    __asm;

    pop iy ; // ret address first because __z88dk_callee convention

    ; REMARK:
    ; hl = msg send
    ; de = msg receive

    pop hl
    pop de

    push iy ; // ret address last  because __z88dk_callee convention

    call TEOPL_TCPC_ADDRESS

    ret

   __endasm;
}
#endif

void teopl_tcpc_init() {

    g_teopl_tcpc_m4_rom_num = 0xFF;

    cpct_memset(g_teopl_tcpc_server_ip, 0, TEOPL_TCPC_SIZE_SERVER_IP);
    cpct_memcpy(g_teopl_tcpc_server_ip, "18.218.56.216", 13);
}

void teopl_tcpc_send_receive(u8* msg_send, u8* msg_receive) {

    // REMARK: 16 is enough although 12 is used (1 + 4 + 7) = (m4rom + ip + password)
    u8 buffer[16]; // FIXME: update this to 12 (but also update asm code)
    u32 ip[4];
    u8 i;
    u8 str_len = 0;
    u8* ptr = g_teopl_tcpc_server_ip;

    // ---

    for (i = 0; i < 4; ++i) {
        ip[i] = teopl_util_string_to_u32(ptr, &str_len, 0);
        ptr += (str_len + 1);
    }

    // prepare message buffers
    // ---

    buffer[0] = g_teopl_tcpc_m4_rom_num;
    buffer[4] = (u8)(ip[0]);
    buffer[3] = (u8)(ip[1]);
    buffer[2] = (u8)(ip[2]);
    buffer[1] = (u8)(ip[3]);

    cpct_memcpy(buffer + 5, msg_send, TEOPL_LEVEL_PASSWORD_LENGTH);

    // disable interrupt and sound
    // REMARK: must be manually enabled
    teopl_interrupt_and_sound(0);

    // send / receive
    // ---

    teopl_pool_extract(TEOPL_POOL_TYPE_TCPC, 0);

    cpct_enableUpperROM();
    _send_receive(buffer, msg_receive);
    cpct_disableUpperROM();

    teopl_interrupt_and_sound(1);

    // remember m4 rom found
    g_teopl_tcpc_m4_rom_num = buffer[0];
}

#endif
