; TCP Echo client example for M4 Board
; Requires firmware v1.0.9b7 upwards
; Duke 2016
; https://github.com/M4Duke/M4examples/blob/master/tcp.s
;
; modified by Naad (@teopl) for "LUDIC - BREAK THE LOOP"

nolist

DATAPORT        equ &FE00
ACKPORT         equ &FC00			
C_NETSOCKET     equ &4331
C_NETCONNECT    equ &4332
C_NETCLOSE      equ &4333
C_NETSEND       equ &4334
C_NETRECV       equ &4335
C_NETHOSTIP     equ &4336

; REMARK:
; hl = msg send
; de = msg receive

tcpc_start:
            
            ; INPUT PARAMETER STRING: 'RXYZW' + 'BEEF123'
            ; ===

            ; R = rom number (255 if unknown)
            ; X,Y,Z,W = ip address
            ; BEEF123 = level password (7 bytes)

            ; 0. MESSAGE TO RETURN ADDRESS
            ; ===

            ld (address_msg_receive), de

            ; 1. M4 rom number (input) - 255 if unknown, else one previously found
            ; ===

            ld a, (hl)
            ld (m4_rom_num), a

            ; 2. M4 rom number (output) - to write back to C, when found
            ; ===

            ld (address_m4_rom_num), hl

            ; 3. IP address 
            ; ===

            inc hl
            ld a, (hl)
            ld (ip_addr), a

            inc hl
            ld a, (hl)
            ld (ip_addr+1), a

            inc hl
            ld a, (hl)
            ld (ip_addr+2), a

            inc hl
            ld a, (hl)
            ld (ip_addr+3), a

            ; 4. MESSAGE TO SEND (LEVEL PASSWORD)
            ; ===

            inc hl
            ld (address_msg_send), hl

            ; ===

            push	iy
            push	ix

            ld	a, (m4_rom_num)
            cp	&FF
            jr z, find_rom

            ; if rom number already found, just enable it and continue
            ld	    a, (m4_rom_num)
            ld      c, a
            call	&B90F		; system/interrupt friendly
            jr and_continue

find_rom:

			call find_m4_rom	    ; find rom (only first run)
						            ; should add version check too and make sure its v1.0.9

            ; write rom number back
            ld hl, (address_m4_rom_num)
            ld a, (m4_rom_num)
            ld (hl), a

and_continue:

			call nz,tcpclient
			
			pop	ix
			pop	iy
			ret
	
tcpclient:	

            ld	hl,&FF02	; get response buffer address
			ld	e,(hl)
			inc	hl
			ld	d,(hl)
			push	de
			pop	iy
			
			; get a socket
			
			ld	hl,cmdsocket
			call	sendcmd
			ld	a,(iy+3)
			cp	255
			ret	z
			
			; store socket in predefined packets
			
			ld	(csocket),a
			ld	(clsocket),a
			ld	(rsocket),a
			ld	(sendsock),a
			
			
			; multiply by 16 and add to socket status buffer
			
			sla	a
			sla	a
			sla	a
			sla	a
			
			ld	hl,&FF06	; get sock info
			ld	e,(hl)
			inc	hl
			ld	d,(hl)
			ld	l,a
			ld	h,0
			add	hl,de	; sockinfo + (socket*4)
			push	hl
			pop	ix		; ix ptr to current socket status
			
			; connect to server

			ld	hl,cmdconnect
			call	sendcmd
			ld	a,(iy+3)
			cp	255
			jp	z,exit_close

wait_connect:

			ld	a,(ix)			; get socket status  (0 ==IDLE (OK), 1 == connect in progress, 2 == send in progress)

			cp	1				; connect in progress?
			jr	z,wait_connect

			cp	0
			jr	z,connect_ok

			call	disp_error	
			jp	exit_close

connect_ok:	

mainloop:	
            ld hl, (address_msg_send)
            ld de, sendtext
            ld bc, 7
            ldir

            ld hl,sendtext
            ld bc, 7
            ld (sendsize),bc

wait_send:	

            ld	a,(ix)
			cp	2			; send in progress?
			jr	z,wait_send	; Could do other stuff here!
			cp	0
			call	nz,disp_error	
			
			ld	hl,5
			add	hl,bc
			ex	de,hl
			ld	hl,cmdsend
			ld	(hl),e
			call	sendcmd

						
			; reset size
			ld	hl,0
			ld	(sendsize),hl

wait_recv:			

			ld	bc,325	
			call	recv
			cp	&FF
			jr	z, exit_close	
			cp	3
			jr	z, exit_close
			xor	a
			cp	c
			jr	nz, got_msg
			cp	b
			jr	z,wait_recv	; keep looping, want a reply. Could do other stuff here!
got_msg:	

			; disp received msg
			push	iy
			pop	hl
		
			ld	de,&6
			add	hl,de		; received text pointer

            ; REMARK: change size if TEOPL_LEVEL_LVL_MAX_NUM_BYTES changed on client or server!
            ; TEOPL_LEVEL_LVL_MAX_NUM_BYTES = 18 * 18 + 1 = 325
            ; also replace 325 everywhere !!!

            push bc
            push hl
            push de

            ld bc, 325
            ld de, (address_msg_receive)
            ldir
            
            pop de
            pop hl
            pop bc

exit_close:
			ld	hl,cmdclose
			call	sendcmd			
			ret
			
			; recv tcp data
			; in
			; bc = receive size
			; out
			; a = receive status
			; bc = received size 

			
recv:		; connection still active
			ld	a,(ix)			; 
			cp	3				; socket status  (3 == remote closed connection)
			ret	z
			; check if anything in buffer ?
			ld	a,(ix+2)
			cp	0
			jr	nz,recv_cont
			ld	a,(ix+3)
			cp	0
			jr	nz,recv_cont
			ld	bc,0
			ld	a,1	
			ret
recv_cont:			
			; set receive size
			ld	a,c
			ld	(rsize),a
			ld	a,b
			ld	(rsize+1),a
			
			ld	hl,cmdrecv
			call	sendcmd
			
			ld	bc,0
			ld	a,(iy+3)
			cp	0				; all good ?
			jr	z,recv_ok
			push	af
			call	disp_error
			pop	af
			ld	bc,0
			ret

recv_ok:			
			ld	c,(iy+4)
			ld	b,(iy+5)
			ret
			
			
			;
			; Find M4 ROM location
			;
				
find_m4_rom:

			ld	iy,m4_rom_name	; rom identification line
			ld	d,127		; start looking for from (counting downwards)
			
romloop:		push	de
			;ld	bc,&DF00
			;out	(c),d		; select rom
			ld	c,d
			call	&B90F		; system/interrupt friendly
			ld	a,(&C000)
			cp	1
			jr	nz, not_this_rom
			
			; get rsxcommand_table
			
			ld	a,(&C004)
			ld	l,a
			ld	a,(&C005)
			ld	h,a
			push	iy
			pop	de
cmp_loop:
			ld	a,(de)
			xor	(hl)			; hl points at rom name
			jr	nz, not_this_rom
			ld	a,(de)
			inc	hl
			inc	de
			and	&80
			jr	z,cmp_loop
			
			; rom found, store the rom number
			
			pop	de			;  rom number
			ld 	a,d
			ld	(m4_rom_num),a
			ret
			
not_this_rom:
			pop	de
			dec	d
			jr	nz,romloop

			ld	a,255		; not found!
			ret
			
			;
			; Send command to M4
			;
sendcmd:
			ld	bc,&FE00
			ld	d,(hl)
			inc	d
sendloop:	
            inc	b
			outi
			dec	d
			jr	nz,sendloop
			ld	bc,&FC00
			out	(c),c
			ret
			
disp_error:
            ret

cmdsocket:	db	5
			dw	C_NETSOCKET
			db	&0,&0,&6		; domain, type, protocol (TCP/ip)

cmdconnect:	db	9	
			dw	C_NETCONNECT
csocket:		db	&0
ip_addr:		db	0,0,0,0		; ip addr
			dw	&1234		; port
cmdsend:		db	0			; we can ignore this byte (part of early design)	
			dw	C_NETSEND
sendsock:		db	0
sendsize:		dw	0			; size
sendtext:		ds	16
			
cmdclose:		db	&03
			dw	C_NETCLOSE
clsocket:		db	&0

cmdrecv:		db	5
			dw	C_NETRECV		; recv
rsocket:		db	&0			; socket
rsize:		dw	0		        ; size
			
m4_rom_name:	db "M4 BOAR",&C4		; D | &80
m4_rom_num:	    db	&FF

; INPUT FROM C FUNCTION ARGUMENTS :
; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

address_msg_send: dw 0

; OUTPUT FROM C FUNCTION ARGUMENTS :
; ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

address_m4_rom_num: dw 0
address_msg_receive: dw 0
